/*
 * pib_control_radio_group.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <stdlib.h>
#include <pib.h>

static PibControlVmt vmt;

static int cmp_val(const void *val, const void *pair)
{
  return *(int *)val-((PibRadioGroupPair *)pair)->val;
}

static int cmp_idc(const void *idc, const void *pair)
{
  return *(int *)idc-((PibRadioGroupPair *)pair)->idc;
}

//#define RADIO_GROUP_CREATE_ORDER
PibControl *pib_control_radio_group_new(PibControlConfig *config,
    PibForm *form, HWND hDlg)
{
#if defined (RADIO_GROUP_CREATE_ORDER) // [
  enum { SIZE=sizeof min[0] };
#endif // ]
  PibControl *control=pib_control_common_new(sizeof control->radio_group,
      &vmt,config,form);
  PibRadioGroupPair *min=config->radio_group.min;
  PibRadioGroupPair *max=config->radio_group.min+config->radio_group.nitems;
  PibRadioGroupPair *cur;
  HWND hPibControl;

  if (!control) {
    DMESSAGE("allocating radio group control");
    goto e_control;
  }

  control->radio_group=config->radio_group;
#if defined (RADIO_GROUP_CREATE_ORDER) // [
  /*
   * we're making sure that the PibRadioGroupPair array is ordered w.r.t. "idc".
   *
   * NOTE: might be *not* useful because as a consequence might shuffle "val".
   */
  qsort(control->radio_group.min,control->radio_group.nitems,SIZE,cmp_idc);
#else // ] [
  for (cur=min;cur<max-1;++cur) {
    if (cur[0].val>=cur[1].val) {
      DVMESSAGE("missing val order: %d>=%d",cur[0].val,cur[1].val);
      goto e_order;
    }
    else if (cur[0].idc>=cur[1].idc) {
      DVMESSAGE("missing idc order: %d>=%d",cur[0].idc,cur[1].idc);
      goto e_order;
    }
  }
#endif // ]

  for (cur=min;cur<max;++cur) {
    hPibControl=GetDlgItem(hDlg,cur->idc);

    SetWindowLongPtrW(
      hPibControl,             // HWND     hWnd,
      GWLP_USERDATA,        // int      nIndex,
      (LONG_PTR)control     // LONG_PTR dwNewLong
    );
  }

  return control;
//cleanup:
#if ! defined (RADIO_GROUP_CREATE_ORDER) // [
e_order:
#endif // ]
  pib_control_common_destroy(control);
e_control:
  return NULL;
}

static void radio_group_destroy(PibControl *control)
{
  pib_control_common_destroy(control);
}

static int radio_group_idc_min(const PibControl *control)
{
  return control->radio_group.min->idc;
}

static int radio_group_idc_max(const PibControl *control)
{
  return (control->radio_group.min+control->radio_group.nitems-1)->idc;
}

static int radio_group_set(PibControl *control, HWND hDlg)
{
  enum { SIZE=sizeof control->radio_group.min[0] };
  size_t nitems=control->radio_group.nitems;
  PibRadioGroupPair *min=control->radio_group.min;
  PibRadioGroupPair *max
      =control->radio_group.min+control->radio_group.nitems-1;
  int val=pib_property_get_int(control->common.property,
      pib_control_cons(control));
  PibRadioGroupPair *pair=bsearch(&val,min,nitems,SIZE,cmp_val);

  if (pair) {
    CheckRadioButton(
      hDlg,         // HWND hDlg,
      min->idc,     // int  nIDFirstButton,
      max->idc,     // int  nIDLastButton,
      pair->idc     // int  nIDCheckButton
    );

    return 0;
  }
  else {
    DVMESSAGE("val:%d not found",val);
    return -1;
  }
}

static int radio_group_get(PibControl *control, HWND hDlg, int idc)
{
  enum { SIZE=sizeof control->radio_group.min[0] };
  size_t nitems=control->radio_group.nitems;
  // get the transient value:
  void *trans=pib_control_trans(control);
  int val=pib_property_get_int(control->common.property,trans);
  PibRadioGroupPair *min=control->radio_group.min;
  PibRadioGroupPair *pair=bsearch(&idc,min,nitems,SIZE,cmp_idc);

DVWRITELN("RADIO GROUP:%p",pair);
  if (pair) {
DVWRITELN("RADIO GROUP:%d %d",pair->val,val);
    if (pair->val!=val) {
      // in case the transient value has changed update it:
      pib_property_set_int(control->common.property,trans,pair->val);

      // let the client know (and let it decide wether the consolidated
      // value is updated too:)
      pib_control_changed(control,hDlg);
    }

    return 0;
  }
  else {
    DMESSAGE("not found");
    return -1;
  }
}

static void radio_group_notify(PibControl *control, HWND hDlg, UINT uMsg,
    WPARAM wParam, LPARAM lParam)
{
  switch (uMsg) {
  case WM_COMMAND:
    /*
     * in case we're notified with a WM_COMMAND we need to read-out the
     * radio group.
     */
    radio_group_get(control,hDlg,LOWORD(wParam));
    break;
  default:
    break;
  }
}

static void radio_group_post(PibControl *control, HWND hDlg, UINT uMsg,
    WPARAM wParam, LPARAM lParam)
{
  switch (uMsg) {
  case WM_PIB_IPC:
    switch (wParam) {
    case PIB_REFRESH:
      radio_group_set(control,hDlg);
      break;
    default:
      break;
    }

    break;
  default:
    break;
  }
}

static PibControlVmt vmt={
#if defined (DEBUG) // [
  .id=__FILE__,
#endif // ]
  .destroy=radio_group_destroy,
  .idc_min=radio_group_idc_min,
  .idc_max=radio_group_idc_max,
  .get=radio_group_get,
  .notify=radio_group_notify,
  .post=radio_group_post,
};
