/*
 * pib_form.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <stdlib.h>
#include <pib.h>

static int cmp_idc(const void *idc, const void *control)
{
  int min=pib_control_idc_min(*(const PibControl **)control);
  int max=pib_control_idc_max(*(const PibControl **)control);

  return *(const int*)idc<min?-1:max<*(const int*)idc?1:0;
}

int pib_form_common_create(PibFormInit *init, HWND hDlg)
{
  enum { SIZE=sizeof init->form->min[0] };
  PibControlConfig *config=init->form_vmt->config;
  PibForm *form=init->form;
  PibControlConfig *cur;

  form->vmt=init->form_vmt;
  form->properties=init->properties;
  form->data=init->data;
  form->min=calloc(pib_form_ncontrols(form),sizeof form->min[0]);

  if (!form->min) {
    DMESSAGE("allocating form");
    goto e_form;
  }

  form->max=form->min;

  for (cur=config;cur<config+pib_form_ncontrols(form);++cur) {
    *form->max=cur->control_new(cur,form,hDlg);

    if (!*form->max) {
      DMESSAGE("creating control");
      goto e_control;
    }

    ++form->max;
  }

  /*
   * we're making sure that the PibControl array is ordered w.r.t. "idc".
   */
  qsort(form->min,pib_form_ncontrols(form),SIZE,cmp_idc);

  return 0;
e_control:
  //TODO: destroy controls.
  free(form->min);
e_form:
  return -1;
}

void pib_form_common_destroy(PibForm *form)
{
  if (form->min) {
    while (form->min<form->max) {
      --form->max;
      pib_control_destroy(*form->max);
    }
  
    free(form->min);
  }
}

int pib_form_refresh(PibForm *form, HWND hDlg)
{
  PibControl **control;

  if (form->min) {
    for (control=form->min;control<form->max;++control) {
      if (!*control) {
        DMESSAGE("uninitialized control");
        continue;
      }

      pib_control_refresh(*control,hDlg);
    }
  }

  return 0;
}

/*
 * intended to be called from some event handler when the "idc" is already
 * known and the accociated "val" has to be set in "options".
 */
int pib_form_get(PibForm *form, HWND hDlg, int idc)
{
  enum { SIZE=sizeof form->min[0] };
  PibControl **min=form->min;
  PibControl **max=form->max;
  size_t nitems=max-min;
  PibControl **control=bsearch(&idc,min,nitems,SIZE,cmp_idc);

  if (!control) {
    DVMESSAGE("not found: %d",idc);
    goto e_control;
  }
  else if (pib_control_get(*control,hDlg,idc)<0) {
    DMESSAGE("setting control");
    goto e_control;
  }

  return 0;
e_control:
  return -1;
}

void pib_form_post(PibForm *form, HWND hDlg, UINT uMsg, WPARAM wParam,
		LPARAM lParam)
{
  PibControl **control;

  for (control=form->min;control<form->max;++control)
    pib_control_post(*control,uMsg,wParam,lParam,hDlg);
}
