/*
 * pib_thread_common.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <pib.h>

static DWORD WINAPI pib_thread_proc(LPVOID lpParameter);

int pib_thread_common_create(PibThread *thread, void *data)
{
  thread->request.next=1;
  thread->data=data;
  thread->head=NULL;
  thread->tail=NULL;

  ////////
  thread->hMutex=CreateMutexW(
    NULL,             // LPSECURITY_ATTRIBUTES lpMutexAttributes,
    FALSE,            // BOOL                  bInitialOwner,
    NULL              // LPCWSTR               lpName
  );

  if (!thread->hMutex) {
    DMESSAGE("creating sync mutex");
    goto e_sync_mutex;
  }

  ////////
  thread->hEvent=CreateEventW(
    NULL,             // LPSECURITY_ATTRIBUTES lpEventAttributes,
    FALSE,            // BOOL                  bManualReset,
    FALSE,            // BOOL                  bInitialState,
    NULL              // LPCWSTR               lpName
  );

  if (!thread->hEvent) {
    DMESSAGE("creating event");
    goto e_event;
  }

  ////////
  thread->hThread=CreateThread(
    NULL,             // LPSECURITY_ATTRIBUTES   lpThreadAttributes,
    0lu,              // SIZE_T                  dwStackSize,
    pib_thread_proc,  // LPTHREAD_START_ROUTINE  lpStartAddress,
    thread,           // __drv_aliasesMem LPVOID lpParameter,
    0lu,              // DWORD                   dwCreationFlags,
    NULL              // LPDWORD                 lpThreadId
  );

  if (!thread->hThread) {
    DMESSAGE("creating sync thread");
    goto e_sync_thread;
  }

  WaitForSingleObject(thread->hEvent,INFINITE);
  return 0;
//cleanup:
  CloseHandle(thread->hThread);
e_sync_thread:
  CloseHandle(thread->hEvent);
e_event:
  CloseHandle(thread->hMutex);
e_sync_mutex:
  return -1;
}

void pib_thread_send(PibThread *thread, PibThreadRequest *request)
{
  WaitForSingleObject(thread->hMutex,INFINITE); // {
  request->prev=thread->tail;
  request->next=NULL;

  if (thread->tail)
    thread->tail->next=request;
  else
    thread->head=request;

  thread->tail=request;
DVWRITELN("### send:\"%s\"",pib_thread_request_id(request));
  SetEvent(thread->hEvent);
  ReleaseMutex(thread->hMutex); // }
}

PibThreadRequest *pib_thread_receive(PibThread *thread)
{
  PibThreadRequest *request;

  WaitForSingleObject(thread->hMutex,INFINITE); // {

  while (!thread->head) {
    ReleaseMutex(thread->hMutex); // {
    WaitForSingleObject(thread->hEvent,INFINITE);
    WaitForSingleObject(thread->hMutex,INFINITE); // }
  }

  request=thread->head;
  thread->head=request->next;

  if (!thread->head)
    thread->tail=NULL;

  request->next=NULL;
DVWRITELN("### received:\"%s\"",pib_thread_request_id(request));
  ReleaseMutex(thread->hMutex); // }

  return request;
}

void pib_thread_common_destroy(PibThread *thread)
{
  PibThreadRequest *request;

  if (thread->hThread) {
    request=pib_thread_request_stop_new(thread);

    if (request) {
      pib_thread_send(thread,request);
      WaitForSingleObject(thread->hThread,INFINITE);
      CloseHandle(thread->hThread);
    }
    else
      DMESSAGE("creation stop request");
  }

  if (thread->hEvent)
    CloseHandle(thread->hEvent);

  if (thread->hMutex)
    CloseHandle(thread->hMutex);
}

static DWORD WINAPI pib_thread_proc(LPVOID lpParameter)
{
  PibThread *thread=lpParameter;
  PibThreadRequest *request;

  SetEvent(thread->hEvent);

  while (thread->request.next) {
//DWRITELN(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
    request=pib_thread_receive(thread);
//DVWRITELN(">>> id:\"%s\"",pib_thread_request_id(request));
    pib_thread_request_execute(request);
    pib_thread_request_destroy(request);
//DVWRITELN("<<< next:%d",thread->request.next);
//DWRITELN("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
  }
//DWRITELN("==========================================");

  return 0;
}
