/*
 * pib_widget_pezold.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */

#include <pib_widget.h>

#if defined (PIB_WIDGET_PETZOLD1) // [
/*
 * adapted from "JUSTIFY1.C" in Charles Petzold: "Programming Windows",
 * Fifth Edition.
 */
void pib_widget_petzold_justify(HDC hDc, LPCWSTR text, SIZE *extent)
{
  int count ;
  LPCWSTR begin, end;
  SIZE size ;
     
  extent->cy = 0;

  // for each text line
  do {
    // initialize number of spaces in line
    count = 0 ;

    // skip over leading spaces
    while (' '==*text)
      text++ ;

    // set pointer to char at beginning of line
    begin = text ;
          
    // until the line is known
    do {
      // set pointer to char at end of line
      end = text ;

      // skip to next space 
      while ('\0'!=*text&&' '!=*text++)
        ;

      if ('\0'==*text)
        break ;

      // after each space encountered, calculate extents
      count++ ;

      GetTextExtentPoint32W(
        hDc,              // HDC     hdc,
        begin,            // LPCWSTR lpString,
        text-begin-1,     // int     c,
        &size             // LPSIZE  psizl
      );
    } while (size.cx<extent->cx) ;
          
    // discount last space at end of line
    --count;
          
    // eliminate trailing spaces
    while (' '==end[- 1]) {
      --end;
      --count;
    }

    // if end of text and no space characters, set end to end
    if ('\0'==*text||count<=0)
      end=text ;

    GetTextExtentPoint32W(
      hDc,                // HDC     hdc,
      begin,              // LPCWSTR lpString,
      end-begin,          // int     c,
      &size               // LPSIZE  psizl
    );

    extent->cy+=size.cy;
    text=end ;
  } while (*text);
}
#elif defined (PIB_WIDGET_INTITALIZER_ARRANGE) // ] [
void ClearRect(HDC hDc, RECT *rect)
{
  HBRUSH hBrushFill=CreateSolidBrush(RGB(255,255,255)); // {
  HBRUSH hBrushOld=(HBRUSH)SelectObject(hDc,hBrushFill); // {

  Rectangle(
    hDc,              // HDC hdc,
    rect->left,       // int left,
    rect->top,        // int top,
    rect->right,      // int right,
    rect->bottom      // int bottom
  );

  SelectObject(hDc,hBrushOld); // }
  DeleteObject(hBrushFill); // }
}

//#define PIB_WIDGET_PETZOLD_CREATE_DUMP
int pib_widget_petzold_create(PibWidgetPetzold *petzold, HWND hDlg,
    HFONT hFont)
{
#if defined (PIB_WIDGET_INTITALIZER_ARRANGE_CORRECT) // [
  int cx,cy;
#if 0 // [
  HBRUSH hBrushFill,hBrushOld;
#else // ] [
  RECT rect;
#endif // ]
#if defined (PIB_WIDGET_PETZOLD_CREATE_DUMP) // [
  int x,y;
  COLORREF pixel;
#endif // ]
#endif // ]

  petzold->hDlg=hDlg;

  /*
   * After painting with a common DC, the ReleaseDC function must be called
   * to release the DC.
   */
  petzold->hDc=GetDC(hDlg);

  if (!petzold->hDc) {
    DMESSAGE("getting device context");
    goto e_hdc;
  }

  petzold->hFont=SelectObject(petzold->hDc,hFont);

#if defined (PIB_WIDGET_INTITALIZER_ARRANGE_CORRECT) // [
  cx=GetDeviceCaps(petzold->hDc,HORZRES);
  cy=GetDeviceCaps(petzold->hDc,VERTRES);

  /*
   * The CreateCompatibleBitmap function creates a bitmap compatible with
   * the device that is associated with the specified device context.
   *
   * If an application sets the nWidth or nHeight parameters to zero,
   * CreateCompatibleBitmap returns the handle to a 1-by-1 pixel, monochrome
   * bitmap.
   *
   * When you no longer need the bitmap, call the DeleteObject function
   * to delete it.
   */
  petzold->compatible.hBitmap=CreateCompatibleBitmap(
    petzold->hDc,             // HDC hdc,
    cx,                       // int cx,
    cy                        // int cy
  );

  if (!petzold->compatible.hBitmap) {
    DMESSAGE("creating compatible bitmap");
    goto e_compatible_hbitmap;
  }

  /*
   * The CreateCompatibleDC function creates a memory device context (DC)
   * compatible with the specified device.
   *
   * When you no longer need the memory DC, call the DeleteDC function.
   */
  petzold->compatible.hDc=CreateCompatibleDC(petzold->hDc);

  if (!petzold->compatible.hDc) {
    DMESSAGE("getting hDc");
    goto e_compatible_hdc;
  }

#if 0 // [
  /*
   * The BitBlt function performs a bit-block transfer of the color data
   * corresponding to a rectangle of pixels from the specified source device
   * context into a destination device context.
   */
  BitBlt(
    petzold->compatible.hDc,  // HDC   hdc,
    0,                        // int   x,
    0,                        // int   y,
    cx,                       // int   cx,
    cy,                       // int   cy,
    petzold->hDc,             // HDC   hdcSrc,
    0,                        // int   x1,
    0,                        // int   y1,
    SRCCOPY                   // DWORD rop
  );
#endif // ]

  SelectObject(petzold->compatible.hDc,petzold->compatible.hBitmap);

#if 0 // [
  hBrushFill=CreateSolidBrush(RGB(255,255,255));
  hBrushOld=(HBRUSH)SelectObject(petzold->compatible.hDc,hBrushFill);

  Rectangle(
    petzold->compatible.hDc,// HDC hdc,
    0,// int left,
    0,// int top,
    cx,// int right,
    cy// int bottom
  );

  SelectObject(petzold->compatible.hDc,hBrushOld);
#else // ] [
  rect.left=0;
  rect.top=0;
  rect.right=cx;
  rect.bottom=cy;
  ClearRect(petzold->compatible.hDc,&rect);
#endif // ]

#if defined (PIB_WIDGET_PETZOLD_CREATE_DUMP) // [
for (x=0;x<cx;++x) {
  for (y=0;y<cy;++y) {
      pixel=GetPixel(petzold->compatible.hDc,x,y);
DVWRITELNW("(%d,%d) %0lx",x,y,pixel);
  }
}
#endif // ]
#endif // ]

  return 0;
//cleanup:
#if defined (PIB_WIDGET_INTITALIZER_ARRANGE_CORRECT) // [
  DeleteDC(petzold->compatible.hDc);
e_compatible_hdc:
  DeleteObject(petzold->compatible.hBitmap);
e_compatible_hbitmap:
#endif // ]
  ReleaseDC(petzold->hDlg,petzold->hDc);
e_hdc:
  return -1;
}

void pib_widget_petzold_destroy(PibWidgetPetzold *petzold)
{
#if defined (PIB_WIDGET_INTITALIZER_ARRANGE_CORRECT) // [
  DeleteDC(petzold->compatible.hDc);
  DeleteObject(petzold->compatible.hBitmap);
#endif // ]
  SelectObject(petzold->hDc,petzold->hFont);
  ReleaseDC(petzold->hDlg,petzold->hDc);
}

#if 0 // [
/*
 * KnowledgeBase Archive
 * An Archive of Early Microsoft KnowledgeBase Articles
 *
 * Q94646: INFO: Using GetCharABCWidths() to Calculate Text Extents
 *   https://jeffpar.github.io/kbarchive/kb/094/Q94646/
 */
BOOL NEAR PASCAL IsBoldSimulated( HDC hDC )
{

  TEXTMETRIC                 tm ;
  OUTLINETEXTMETRIC          otm ;

  GetTextMetrics( hDC, &tm ) ;

  if (tm.tmWeight <= FW_NORMAL)
     return FALSE ;

  otm.otmSize = sizeof( OUTLINETEXTMETRIC ) ;

  /* Call GOTM() with the pointer to the buffer.  It will
   * fill in the buffer.
   */ 
  if (!GetOutlineTextMetrics( hDC, otm.otmSize, &otm ))
     return FALSE ;  // Probably not a TrueType font

       /* See the documentation on OUTLINETEXTMETRIC for a
        * description of the otmfsSelection member.
        */ 
  if (!(otm.otmfsSelection & 0x0020)) // bit 5 means bold
        return TRUE ;


#if 1 // [
  return FALSE;
#endif // ]
} /* IsBoldSimulated()  */ 

/****************************************************************
*  UINT WINAPI GetTextExtentABCPoint( HDC hDC, LPSTR lpsz,
*                                     int cb, LPSIZE lpSize )
*
*    Just like GetTextExtentPoint() but only works on TT fonts
*    and uses GetCharABCWidths().
*
*    Unlike GetTextExtent() and GetTextExtentPoint() which
*    return a BOOL, this function returns the absolute value
*    for the "A" value of the first character if it is
*    negative otherwise it returns 0.
*
****************************************************************/ 

UINT WINAPI GetTextExtentABCPointW( HDC hDC, LPCWSTR lpsz,

                                  int cb, LPSIZE lpSize )

{

  int          A, C ;
  UINT         nRet = 0 ;
  static LPABC lpABC ;

#if 1 // [
  lpSize->cx=0;
#endif // ]

  GetTextExtentPointW( hDC, lpsz, cb, lpSize );

  if (lpABC ||
#if 0 // [
  (lpABC = (LPABC)GlobalAllocPtr( GHND, 256*sizeof(ABC) ))
#else // ] [
  (lpABC = (LPABC)calloc(256,sizeof(ABC) ))
#endif // ]
   )
  {
     if (GetCharABCWidths( hDC, 0, 255, lpABC ))
     {
        int         n ;

        /* We now have an array of ABC structures
         */ 
        for ( lpSize->cx = 0, n = 0 ; n < cb ; n++ )
        {
#if 0 // [
           lpSize->cx += (
                          lpABC[(int)lpsz[n]].abcA +
                          lpABC[(int)lpsz[n]].abcB +
                          lpABC[(int)lpsz[n]].abcC ) ;
#else // ] [
           lpSize->cx+=lpABC[(int)lpsz[n]].abcA ;
           lpSize->cx+=lpABC[(int)lpsz[n]].abcB ;
           lpSize->cx+=lpABC[(int)lpsz[n]].abcC ;
#endif // ]
        }

        if( cb )
        {
           A = lpABC[(int)lpsz[0]].abcA;

           if( A > 0 )
           {
              nRet = -A ;
              lpSize->cx -= A;
           }

           C = lpABC[(int)lpsz[cb-1]].abcC ;

           if ( C < 0 )
              lpSize->cx -= C ;    // two minus' make a plus
        }

        /* Bug in Windows 3.1.  If the tt font is embolded then
         * add one to the width for each char.
         */ 
        if ((BOOL)LOWORD( GetVersion() ) == 0x0A03 &&
            IsBoldSimulated( hDC ))
           lpSize->cx += cb ;
     }
     else
        GetTextExtentPointW( hDC, lpsz, cb, lpSize ) ;

     return nRet  ;
  }
  return nRet ;

} /* GetTextExtentABCPoint()  */ 
#endif // ]

void pib_widget_petzold_extent(PibWidgetPetzold *petzold, LPCWSTR pText,
    SIZE *extent)
{
#if defined (PIB_WIDGET_INTITALIZER_ARRANGE_CORRECT) // [
  int len=wcslen(pText);
  SIZE last;
  RECT rect UNUSED;
  int x,y;
  COLORREF pixel;
  //HBRUSH hBrushFill,hBrushOld;
#endif // ]

#if defined (PIB_WIDGET_INTITALIZER_ARRANGE_CORRECT) // [

  /*
   * Width of Text in Italic Font
   *   https://www.codeproject.com/Articles/14915/Width-of-Text-in-Italic-Font
   *
   * 1) Calculate the width of the text, by using the classic method.
   */
#if 0 // [
  GetTextExtentPoint32W(
    petzold->hDc,       // HDC     hdc,
    pText,              // LPCWSTR lpString,
    len,                // int     c,
    extent              // LPSIZE  psizl
  );
#else // ] [
  GetTextExtentABCPointW(petzold->hDc,pText,wcslen(pText),extent);
DVWRITELNW("\"%s\" %ld",pText,extent->cx);
#endif // ]
DVWRITELNW("(1) \"%s\": %d %d",pText,extent->cx,extent->cy);

  /*
   * 2) Calculate the width of the last character, as suggested by 'oupoi'.
   */
  GetTextExtentPoint32W(
    petzold->compatible.hDc,       // HDC     hdc,
    pText+len-1,        // LPCWSTR lpString,
    1,                  // int     c,
    &last               // LPSIZE  psizl
  );

  /*
   * 3) Set a bounding rectangle wide enough to fit the painted text.
   */
  rect.left=0;
  rect.top=0;
  rect.right=extent->cx+last.cx;
  rect.bottom=extent->cy;

#if 0 // [
  for (x=rect.right-1;0<=x;--x) {
    for (y=0;y<rect.bottom;++y) {
      pixel=GetPixel(petzold->hDc,x,y);
DVWRITELNW("(2) \"%s\": (%d,%d) %0lx",pText,x,y,pixel);
    }
  }
#endif // ]

  /*
   * 4) Paint the text.
   */
  DrawTextW(
    petzold->compatible.hDc,       // HDC     hdc,
    pText,              // LPCWSTR lpchText,
    -1,                 // int     cchText,
    &rect,              // LPRECT  lprc,
    DT_LEFT|DT_TOP|DT_SINGLELINE|DT_NOPREFIX
                        // UINT    format
  );

  /*
   * 5) Then, scan the colour of the pixels in the memory device context,
   *    from right-to-left
   */
#if 1 // [
  for (x=rect.right-1;0<=x;--x) {
    for (y=0;y<rect.bottom;++y) {
      pixel=GetPixel(petzold->compatible.hDc,x,y);
DVWRITELNW("(2) \"%s\": %0lx",pText,pixel);

      if (RGB(255,255,255)!=pixel) {
extent->cx=x+1;
//extent->cx/=2;
//extent->cx=MulDiv(8,x+1,52);
DVWRITELNW("(3) \"%s\": %d %d",pText,extent->cx,extent->cy);
#if 0 // [
BitBlt(
  petzold->compatible.hDc,
  rect.left,
  rect.top,
  rect.right-rect.left,
  rect.bottom-rect.top,
  petzold->hDc,
  rect.left,
  rect.top,
  SRCCOPY
);
#else // ] [
#if 0 // [
  hBrushFill=CreateSolidBrush(RGB(255,255,255));
  hBrushOld=(HBRUSH)SelectObject(petzold->compatible.hDc,hBrushFill);

  Rectangle(
    petzold->compatible.hDc,// HDC hdc,
    rect.left,// int left,
    rect.top,// int top,
    rect.right,// int right,
    rect.bottom// int bottom
  );

  SelectObject(petzold->compatible.hDc,hBrushOld);
#else // ] [
  ClearRect(petzold->compatible.hDc,&rect);
#endif // ]
#endif // ]

        return;
      }
    }
  }
#endif // ]
#else // ] [
  GetTextExtentABCPointW(petzold->hDc,pText,wcslen(pText),extent);
DVWRITELNW("\"%s\" %ld",pText,extent->cx);
#endif // ]
}
#elif defined (PIB_WIDGET_PETZOLD2) // ] [
#define PIB_WIDGET_PETZOLD2_ABC
/*
 * adapted from "JUSTIFY2.C" in Charles Petzold: "Programming Windows",
 * Fifth Edition, p. 830 ff.
 */
#if 0 // [
#define LASTCHAR 127     // Last character code used in text
#endif // ]

/*----------------------------------------------------------------------
   GetCharDesignWidths:  Gets character widths for font as large as the
                         original design size
  ----------------------------------------------------------------------*/

#if defined (PIB_WIDGET_PETZOLD2_ABC) // [
static UINT GetCharDesignWidths (PibWidgetPetzold *petzold, UINT uFirst,
    UINT uLast, ABCFLOAT * pAbc)
#elif defined (PIB_WIDGET_PETZOLD2) // ] [
static UINT GetCharDesignWidths (PibWidgetPetzold *petzold, UINT uFirst,
    UINT uLast, int * piWidths)
#else // ] [
UINT GetCharDesignWidths (HDC hDc, UINT uFirst, UINT uLast, int * piWidths)
#endif // ]
{
     HFONT             hFont, hFontDesign ;
     LOGFONTW          lf ;
     int code;
     OUTLINETEXTMETRICW otm ;
#if ! defined (PIB_WIDGET_PETZOLD2) \
    && ! defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     SIZE size;
#endif // ]

     /*
      * The GetCurrentObject function retrieves a handle to an object of
      * the specified type that has been selected into the specified device
      * context (DC).
      */
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
//#if defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
//     hFont = (HFONT)SendMessageW(petzold->hDlg,WM_GETFONT,0,0);
//#else // ] [
#if defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
DVWRITELN("hDlg:%p hDc:%p",petzold->hDlg,petzold->hDc);
#endif // ]
    GetTextFaceW(
      petzold->hDc,     // HDC    hdc,
      (sizeof lf.lfFaceName)/(sizeof lf.lfFaceName[0]),
                        // int    c,
      lf.lfFaceName     // LPWSTR lpName
    );
DVWRITELNW("\"%s\"",lf.lfFaceName);

     hFont = GetCurrentObject(
        petzold->hDc,               // HDC  hdc,
        OBJ_FONT                    // UINT type
     ) ;
//#endif // ]
#else // ] [
     hFont = GetCurrentObject(
        hDc,                        // HDC  hdc,
        OBJ_FONT                    // UINT type
     ) ;
#endif // ]

     /*
      * The GetObject function retrieves information for the specified
      * graphics object.
      */
memset(&lf,0,sizeof lf);
     code=GetObjectW (
        hFont,                      // HANDLE h,
        sizeof (LOGFONTW),          // int    c,
        &lf                         // LPVOID pv
     ) ;
DVWRITELNW("\"%s\": %d (%lu)",lf.lfFaceName,code,*lf.lfFaceName);
if (!code||!*lf.lfFaceName||254<*lf.lfFaceName) {
  DVMESSAGE("something went wrong: %lu",*lf.lfFaceName);
  return 0;
}

          // Get outline text metrics (we'll only be using a field that is
          //   independent of the DC the font is selected into)

     otm.otmSize = sizeof (OUTLINETEXTMETRIC) ;

     /*
      * The GetOutlineTextMetrics function retrieves text metrics for
      * TrueType fonts.
      */
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     GetOutlineTextMetricsW(
        petzold->hDc,               // HDC                  hdc,
        sizeof (OUTLINETEXTMETRIC), // UINT                 cjCopy,
        &otm                        // LPOUTLINETEXTMETRICA potm
     ) ;
#else // ] [
     GetOutlineTextMetricsW(
        hDc,                        // HDC                  hdc,
        sizeof (OUTLINETEXTMETRIC), // UINT                 cjCopy,
        &otm                        // LPOUTLINETEXTMETRICA potm
     ) ;
#endif // ]
if (!otm.otmEMSquare) {
  DMESSAGE("something went wrong");
  exit(1);
}

          // Create a new font based on the design size

     /*
      * lfHeight < 0: The font mapper transforms this value into device
      * units and matches its absolute value against the character height
      * of the available fonts.
      */
     lf.lfHeight = - (int) otm.otmEMSquare ;
     lf.lfWidth  = 0 ;

     /*
      * The CreateFontIndirect function creates a logical font that has the
      * specified characteristics. The font can subsequently be selected as
      * the current font for any device context.
      */
     hFontDesign = CreateFontIndirectW (
        &lf                         // const LOGFONTW *lplf
     ) ;

          // Select the font into the DC and get the character widths

     /*
      * The SaveDC function saves the current state of the specified device
      * context (DC) by copying data describing selected objects and graphic
      * modes (such as the bitmap, brush, palette, font, pen, region, drawing
      * mode, and mapping mode) to a context stack.
      */ 
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     SaveDC (petzold->hDc) ;
#else // ] [
     SaveDC (hDc) ;
#endif // ]

     /*
      * The SetMapMode function sets the mapping mode of the specified
      * device context. The mapping mode defines the unit of measure used
      * to transform page-space units into device-space units, and also
      * defines the orientation of the device's x and y axes.
      */
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     SetMapMode (petzold->hDc, MM_TEXT) ;
#else // ] [
     SetMapMode (hDc, MM_TEXT) ;
#endif // ]

     /*
      * The SelectObject function selects an object into the specified
      * device context (DC). The new object replaces the previous object
      * of the same type.
      */
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     SelectObject (petzold->hDc, hFontDesign) ;
#else // ] [
     SelectObject (hDc, hFontDesign) ;
#endif // ]

     /*
      * The GetCharWidth function retrieves the widths, in logical
      * coordinates, of consecutive characters in a specified range
      * from the current font.
      *
      * NOTE: This function is provided only for compatibility with
      *   16-bit versions of Windows. Applications should call the
      *   GetCharWidth32 function, which provides more accurate results.
      */
#if 0 // [
#if defined (PIB_WIDGET_PETZOLD2) // [
     GetCharWidthW (
        petzold->hDc,   // HDC   hdc,
        uFirst,         // UINT  iFirst,
        uLast,          // UINT  iLast,
        piWidths        // LPINT lpBuffer
     ) ;
#if 1 // [
    // Call this function just once to get size.cy (font height)
     GetTextExtentPoint32W(
        petzold->hDc,   // HDC     hdc,
        L"X",          // LPCWSTR lpString,
        1,              // int     c,
        &size           // LPSIZE  psizl
     ) ;

     petzold->cy=size.cy;
DVWRITELN("cy:%.0f",petzold->cy);
#endif // ]

#else // ] [
     GetCharWidthW (
        petzold->hDc,       // HDC   hdc,
        uFirst,         // UINT  iFirst,
        uLast,          // UINT  iLast,
        piWidths        // LPINT lpBuffer
     ) ;
#endif // ]
#elif ! defined (PIB_WIDGET_PETZOLD2_ABC) // ] [
     /*
      * GetCharWidth32 cannot be used on TrueType fonts. To retrieve
      * character widths for TrueType fonts, use GetCharABCWidths.
      */
     GetCharWidth32W(
        petzold->hDc,       // HDC   hdc,
        uFirst,             // UINT  iFirst,
        uLast,              // UINT  iLast,
        piWidths            // LPINT lpBuffer
     ) ;
#else // ] [
     GetCharABCWidthsFloatW(
        petzold->hDc,       // HDC         hdc,
        uFirst,             // UINT        iFirst,
        uLast,              // UINT        iLast,
        pAbc                // LPABCFLOAT  lpABC
    );
#endif // ]

     /*
      * The SelectObject function selects an object into the specified
      * device context (DC). The new object replaces the previous object
      * of the same type.
      */
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     SelectObject (petzold->hDc, hFont) ;
#else // ] [
     SelectObject (hDc, hFont) ;
#endif // ]

     /*
      * The RestoreDC function restores a device context (DC) to the
      * specified state. The DC is restored by popping state information
      * off a stack created by earlier calls to the SaveDC function.
      */
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     RestoreDC (petzold->hDc, -1) ;
#else // ] [
     RestoreDC (hDc, -1) ;
#endif // ]

          // Clean up

     /*
      * The DeleteObject function deletes a logical pen, brush, font,
      * bitmap, region, or palette, freeing all system resources associated
      * with the object. After the object is deleted, the specified handle
      * is no longer valid.
      */
     DeleteObject (hFontDesign) ;

     return otm.otmEMSquare ;
}

/*---------------------------------------------------------------------
   GetScaledWidths:  Gets floating point character widths for selected
                     font size
  ---------------------------------------------------------------------*/

#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
static void GetScaledWidths (PibWidgetPetzold *petzold)
#else // ] [
void GetScaledWidths (HDC hDc, double * pdWidths)
#endif // ]
{
     double  dScale ;
     HFONT   hFont ;
#if defined (PIB_WIDGET_PETZOLD2_ABC) // [
     ABCFLOAT aiDesignWidths [PIB_WIDGET_PETZOLD_LASTCHAR + 1] ;
#elif defined (PIB_WIDGET_PETZOLD2) // ] [
     int     aiDesignWidths [PIB_WIDGET_PETZOLD_LASTCHAR + 1] ;
#else // ] [
     int     aiDesignWidths [LASTCHAR + 1] ;
#endif // ]
     int     i ;
     LOGFONTW lf ;
     UINT    uEMSquare ;

          // Call function above

#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     uEMSquare = GetCharDesignWidths (petzold, 0, PIB_WIDGET_PETZOLD_LASTCHAR,
        aiDesignWidths) ;
#else // ] [
     uEMSquare = GetCharDesignWidths (hDc, 0, LASTCHAR, aiDesignWidths) ;
#endif // ]

          // Get LOGFONT for current font in device context

     /*
      * The GetCurrentObject function retrieves a handle to an object of
      * the specified type that has been selected into the specified device
      * context (DC).
      */
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     hFont = GetCurrentObject (petzold->hDc, OBJ_FONT) ;
#else // ] [
     hFont = GetCurrentObject (hDc, OBJ_FONT) ;
#endif // ]

     /*
      * The GetObject function retrieves information for the specified
      * graphics object.
      */
     GetObjectW (hFont, sizeof (LOGFONTW), &lf) ;

          // Scale the widths and store as floating point values

if (!uEMSquare) {
  DMESSAGE("something went wrong");
  exit(1);
}

#if 1 // [
     dScale = (double) -lf.lfHeight / (double) uEMSquare ;
#else // ] [
     dScale =(double)lf.lfHeight/uEMSquare ;

     if (dScale<0)
      dScale=-dScale;
#endif // ]

#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     for (i = 0 ; i <= PIB_WIDGET_PETZOLD_LASTCHAR ; i++) {
#if defined (PIB_WIDGET_PETZOLD2_ABC) // [
          double width=0.0;

          //width+=aiDesignWidths[i].abcfA;
          width+=aiDesignWidths[i].abcfB;
          //width+=aiDesignWidths[i].abcfC;
if (aiDesignWidths[i].abcfA<0) DVWRITELN("'%c' -",(char)i);
if (aiDesignWidths[i].abcfC<0) DVWRITELN("'%c' +",(char)i);

          petzold->widths[i] = dScale * width ;
#else // ] [
          petzold->widths[i] = dScale * aiDesignWidths[i] ;
#endif // ]
      }

DVWRITELN("cy:%.0f (%.0f)",petzold->cy,dScale);
#if 0 // [
      petzold->cy*=2.0*dScale;
DVWRITELN("cy:%.0f",petzold->cy);
#endif // ]
#else // ] [
     for (i = 0 ; i <= LASTCHAR ; i++)
          pdWidths[i] = dScale * aiDesignWidths[i] ;
#endif // ]
}

/*--------------------------------------------------------------
   GetTextExtentFloat:  Calculates text width in floating point
  --------------------------------------------------------------*/

#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
static double GetTextExtentFloat (PibWidgetPetzold *petzold, LPCWSTR psText,
    int iCount)
#else // ] [
double GetTextExtentFloat (double * pdWidths, LPCWSTR psText, int iCount)
#endif // ]
{
     double dWidth = 0 ;
     int    i ;

#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     for (i = 0 ; i < iCount ; i++)
          dWidth += petzold->widths [(int)psText[i]] ;
#else // ] [
     for (i = 0 ; i < iCount ; i++)
          dWidth += pdWidths [(int)psText[i]] ;
#endif // ]

     return dWidth ;
}

#if 0 // [
/*
 * KnowledgeBase Archive
 * An Archive of Early Microsoft KnowledgeBase Articles
 *
 * Q94646: INFO: Using GetCharABCWidths() to Calculate Text Extents
 *   https://jeffpar.github.io/kbarchive/kb/094/Q94646/
 */

     BOOL NEAR PASCAL IsBoldSimulated( HDC hDC )
     {

       TEXTMETRIC                 tm ;
       OUTLINETEXTMETRIC          otm ;

       GetTextMetrics( hDC, &tm ) ;

       if (tm.tmWeight <= FW_NORMAL)
          return FALSE ;

       otm.otmSize = sizeof( OUTLINETEXTMETRIC ) ;

       /* Call GOTM() with the pointer to the buffer.  It will
        * fill in the buffer.
        */ 
       if (!GetOutlineTextMetrics( hDC, otm.otmSize, &otm ))
          return FALSE ;  // Probably not a TrueType font

       /* See the documentation on OUTLINETEXTMETRIC for a
        * description of the otmfsSelection member.
        */ 
       if (!(otm.otmfsSelection & 0x0020)) // bit 5 means bold
             return TRUE ;


#if 1 // [
      return FALSE;
#endif // ]
     } /* IsBoldSimulated()  */ 

     /****************************************************************
     *  UINT WINAPI GetTextExtentABCPoint( HDC hDC, LPSTR lpsz,
     *                                     int cb, LPSIZE lpSize )
     *
     *    Just like GetTextExtentPoint() but only works on TT fonts
     *    and uses GetCharABCWidths().
     *
     *    Unlike GetTextExtent() and GetTextExtentPoint() which
     *    return a BOOL, this function returns the absolute value
     *    for the "A" value of the first character if it is
     *    negative otherwise it returns 0.
     *
     ****************************************************************/ 

     UINT WINAPI GetTextExtentABCPointW( HDC hDC, LPCWSTR lpsz,

                                       int cb, LPSIZE lpSize )

     {

       int          A, C ;
       UINT         nRet = 0 ;
       static LPABC lpABC ;

#if 1 // [
       lpSize->cx=0;
#endif // ]

       GetTextExtentPointW( hDC, lpsz, cb, lpSize );

       if (lpABC ||
#if 0 // [
       (lpABC = (LPABC)GlobalAllocPtr( GHND, 256*sizeof(ABC) ))
#else // ] [
       (lpABC = (LPABC)calloc(256,sizeof(ABC) ))
#endif // ]
        )
       {
          if (GetCharABCWidths( hDC, 0, 255, lpABC ))
          {
             int         n ;

             /* We now have an array of ABC structures
              */ 
             for ( lpSize->cx = 0, n = 0 ; n < cb ; n++ )
             {
#if 0 // [
                lpSize->cx += (
                               lpABC[(int)lpsz[n]].abcA +
                               lpABC[(int)lpsz[n]].abcB +
                               lpABC[(int)lpsz[n]].abcC ) ;
#else // ] [
                lpSize->cx +=
                               lpABC[(int)lpsz[n]].abcA ;
                lpSize->cx +=
                               lpABC[(int)lpsz[n]].abcB ;
                lpSize->cx +=
                               lpABC[(int)lpsz[n]].abcC ;
#endif // ]
             }

             if( cb )
             {
                A = lpABC[(int)lpsz[0]].abcA;

                if( A > 0 )
                {
                   nRet = -A ;
                   lpSize->cx -= A;
                }

                C = lpABC[(int)lpsz[cb-1]].abcC ;

                if ( C < 0 )
                   lpSize->cx -= C ;    // two minus' make a plus
             }

             /* Bug in Windows 3.1.  If the tt font is embolded then
              * add one to the width for each char.
              */ 
             if ((BOOL)LOWORD( GetVersion() ) == 0x0A03 &&
                 IsBoldSimulated( hDC ))
                lpSize->cx += cb ;
          }
          else
             GetTextExtentPointW( hDC, lpsz, cb, lpSize ) ;

          return nRet  ;
       }
       return nRet ;

    } /* GetTextExtentABCPoint()  */ 
#endif // ]

void pib_widget_petzold_extent(PibWidgetPetzold *petzold, LPCWSTR pText,
    SIZE *extent)
{
#if defined (PIB_WIDGET_PETZOLD2_CREATE_COMPATIBLE) // [
  int len=wcslen(pText);
  SIZE last;
  RECT rect UNUSED;
  int x,y;
  COLORREF pixel;

  /*
   * Width of Text in Italic Font
   *   https://www.codeproject.com/Articles/14915/Width-of-Text-in-Italic-Font
   *
   * 1) Calculate the width of the text, by using the classic method.
   */
  GetTextExtentPoint32W(
    petzold->hDc,       // HDC     hdc,
    pText,              // LPCWSTR lpString,
    len,                // int     c,
    extent              // LPSIZE  psizl
  );
DVWRITELNW("(1) \"%s\": %d %d",pText,extent->cx,extent->cy);

  /*
   * 2) Calculate the width of the last character, as suggested by 'oupoi'.
   */
  GetTextExtentPoint32W(
    petzold->hDc,       // HDC     hdc,
    pText+len-1,        // LPCWSTR lpString,
    1,                  // int     c,
    &last               // LPSIZE  psizl
  );

  /*
   * 3) Set a bounding rectangle wide enough to fit the painted text.
   */
  rect.left=0;
  rect.top=0;
  rect.right=extent->cx+last.cx;
  rect.bottom=extent->cy;

#if 0 // [
  for (x=rect.right-1;0<=x;--x) {
    for (y=0;y<rect.bottom;++y) {
      pixel=GetPixel(petzold->hDc,x,y);
DVWRITELNW("(2) \"%s\": (%d,%d) %0lx",pText,x,y,pixel);

      if (pixel) {
//extent->cx=x;
//DVWRITELNW("(3) \"%s\": %d %d",pText,extent->cx,extent->cy);
        //goto e_found;
      }
    }
  }
#endif // ]

  /*
   * 4) Paint the text.
   */
  DrawTextW(
    petzold->hDc,       // HDC     hdc,
    pText,              // LPCWSTR lpchText,
    -1,                 // int     cchText,
    &rect,              // LPRECT  lprc,
    DT_LEFT|DT_TOP|DT_SINGLELINE|DT_NOPREFIX
                        // UINT    format
  );

  /*
   * 5) Then, scan the colour of the pixels in the memory device context,
   *    from right-to-left
   */
#if 1 // [
  for (x=rect.right-1;0<=x;--x) {
    for (y=0;y<rect.bottom;++y) {
      pixel=GetPixel(petzold->hDc,x,y);
//DVWRITELNW("(3) \"%s\": %0lx",pText,pixel);

      if (RGB(255,255,255)!=pixel) {
extent->cx=x+1;
extent->cx/=2;
//extent->cx=MulDiv(8,x+1,52);
//DVWRITELNW("(3) \"%s\": %d %d",pText,extent->cx,extent->cy);
BitBlt(
  petzold->hDc,
  rect.left,
  rect.top,
  rect.right-rect.left,
  rect.bottom-rect.top,
  petzold->screen.hDc,
  rect.left,
  rect.top,
  SRCCOPY
);
        return;
      }
    }
  }
#endif // ]
#else // ] [
#if 0 // [
    // Call this function just once to get size.cy (font height)
     GetTextExtentPoint32W(
        petzold->hDc,   // HDC     hdc,
        pText,          // LPCWSTR lpString,
        1,              // int     c,
        extent          // LPSIZE  psizl
     ) ;
#else // ] [
   extent->cy=petzold->cy+0.5;
//DVWRITELN("cy:%ld",extent->cy);
#endif // ]

#if 0 // [
   extent->cx=GetTextExtentFloat(petzold,pText,wcslen(pText)) ;
#else // ] [
     GetTextExtentABCPointW(petzold->hDc,pText,wcslen(pText),extent);
DVWRITELNW("\"%s\" %ld",pText,extent->cx);
#endif // ]
#endif // ]
}

/*------------------------------------------------------------------
   Justify:  Based on design units for screen/printer compatibility
  ------------------------------------------------------------------*/

#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
void pib_widget_petzold_justify(PibWidgetPetzold *petzold, LPCWSTR pText,
    SIZE *extent)
#else // ] [
void pib_widget_petzold_justify(HDC hDc, LPCWSTR pText, SIZE *extent)
#endif // ]
{
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     double dWidth;
#else // ] [
     double dWidth, adWidths[LASTCHAR + 1] ;
#endif // ]
     int    cSpaceChars ;
     LPCWSTR  pBegin, pEnd ;
     SIZE   size ;

          // Fill the adWidths array with floating point character widths

#if ! defined (PIB_WIDGET_PETZOLD2) \
    && ! defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
     GetScaledWidths (hDc, adWidths) ;
#endif // ]

#if 0 // [
          // Call this function just once to get size.cy (font height)

#if defined (PIB_WIDGET_PETZOLD2) // [
     GetTextExtentPoint32W(
        petzold->hDc,   // HDC     hdc,
        pText,          // LPCWSTR lpString,
        1,              // int     c,
        &size           // LPSIZE  psizl
     ) ;
#else // ] [
     GetTextExtentPoint32W(
        hDc,            // HDC     hdc,
        pText,          // LPCWSTR lpString,
        1,              // int     c,
        &size           // LPSIZE  psizl
     ) ;
#endif // ]
#else // ] [
   size.cy=petzold->cy;
#endif // ]
     
     extent->cy=0;

     // for each text line
     do {
          // initialize number of spaces in line
          cSpaceChars = 0 ;

          // skip over leading spaces
          while (*pText == ' ')
               pText++ ;

          // set pointer to char at beginning of line
          pBegin = pText ;
          
          // until the line is known
          do {
               // set pointer to char at end of line
               pEnd = pText ;

               // skip to next space 
               while (*pText != '\0' && *pText++ != ' ')
                  ;

               if (*pText == '\0')
                    break ;

               // after each space encountered, calculate extents
               cSpaceChars++ ;

#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
               dWidth = GetTextExtentFloat (petzold, pBegin, 
                                                      pText - pBegin - 1) ;
#else // ] [
               dWidth = GetTextExtentFloat (adWidths, pBegin, 
                                                      pText - pBegin - 1) ;
#endif // ]
          } while (dWidth < (double) extent->cx) ;
          
          // discount last space at end of line
          cSpaceChars-- ;
          
          // eliminate trailing spaces
          while (*(pEnd - 1) == ' ')
          {
               pEnd-- ;
               cSpaceChars-- ;
          }

          // if end of text and no space characters, set pEnd to end
          if (*pText == '\0' || cSpaceChars <= 0)
               pEnd = pText ;
          
#if defined (PIB_WIDGET_PETZOLD2) \
    || defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
          dWidth = GetTextExtentFloat (petzold, pBegin, pText - pBegin - 1) ;
#else // ] [
          dWidth = GetTextExtentFloat (adWidths, pBegin, pText - pBegin - 1) ;
#endif // ]
          
          extent->cy += size.cy ;
          pText = pEnd ;
     } while (*pText) ;
}

#if defined (PIB_WIDGET_PETZOLD2_CREATE_COMPATIBLE) // [
/*
 * How can I take a screenshot in a windows application?
 *   https://stackoverflow.com/questions/3291167/how-can-i-take-a-screenshot-in-a-windows-application
 */
#endif // ]
#if defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
int pib_widget_petzold_create(PibWidgetPetzold *petzold, HWND hDlg)
#else // ] [
int pib_widget_petzold_create(PibWidgetPetzold *petzold, int point_size,
    const wchar_t *typeface)
#endif // ]
{
#if defined (PIB_WIDGET_PETZOLD2_CREATE_COMPATIBLE) // [
  int cx,cy;

  /*
   * The CreateDC function creates a device context (DC) for a device using
   * the specified name.
   *
   * When you no longer need the DC, call the DeleteDC function.
   */
  petzold->screen.hDc=CreateDCW(L"DISPLAY",NULL,NULL,NULL);

  if (!petzold->screen.hDc) {
    DMESSAGE("creating display device context");
    goto e_screen_hdc;
  }

  cx=GetDeviceCaps(petzold->screen.hDc,HORZRES);
  cy=GetDeviceCaps(petzold->screen.hDc,VERTRES);

  /*
   * The CreateCompatibleBitmap function creates a bitmap compatible with
   * the device that is associated with the specified device context.
   *
   * If an application sets the nWidth or nHeight parameters to zero,
   * CreateCompatibleBitmap returns the handle to a 1-by-1 pixel, monochrome
   * bitmap.
   *
   * When you no longer need the bitmap, call the DeleteObject function
   * to delete it.
   */
  petzold->screen.hBitmap=CreateCompatibleBitmap(
    petzold->screen.hDc,      // HDC hdc,
    cx,                       // int cx,
    cy                        // int cy
  );

  if (!petzold->screen.hBitmap) {
    DMESSAGE("creating compatible bitmap");
    goto e_screen_hbitmap;
  }

  /*
   * The CreateCompatibleDC function creates a memory device context (DC)
   * compatible with the specified device.
   *
   * When you no longer need the memory DC, call the DeleteDC function.
   */
  petzold->hDc=CreateCompatibleDC(petzold->screen.hDc);

  if (!petzold->hDc) {
    DMESSAGE("getting hDc");
    goto e_hdc;
  }

  /*
   * The BitBlt function performs a bit-block transfer of the color data
   * corresponding to a rectangle of pixels from the specified source device
   * context into a destination device context.
   */
  BitBlt(
    petzold->hDc,             // HDC   hdc,
    0,                        // int   x,
    0,                        // int   y,
    cx,                       // int   cx,
    cy,                       // int   cy,
    petzold->screen.hDc,      // HDC   hdcSrc,
    0,                        // int   x1,
    0,                        // int   y1,
    SRCCOPY                   // DWORD rop
  );

  SelectObject(petzold->hDc,petzold->screen.hBitmap);
#elif defined (PIB_WIDGET_INTITALIZER_ARRANGE) // ] [
  HFONT hFont=CreateFontW(
    -GetDeviceCaps(petzold->hDc,LOGPIXELSY)/6,
                                    // int    cHeight,
    0,                              // int    cWidth,
    0,                              // int    cEscapement,
    0,                              // int    cOrientation,
    FW_DONTCARE,                    // int    cWeight,
    FALSE,                          // DWORD  bItalic,
    FALSE,                          // DWORD  bUnderline,
    FALSE,                          // DWORD  bStrikeOut,
    ANSI_CHARSET,                   // DWORD  iCharSet,
    OUT_DEFAULT_PRECIS,             // DWORD  iOutPrecision,
    CLIP_DEFAULT_PRECIS,            // DWORD  iClipPrecision,
    DEFAULT_QUALITY,                // DWORD  iQuality,
    DEFAULT_PITCH|FF_SWISS,         // DWORD  iPitchAndFamily,
    L"Ms Shell Dlg"                 // LPCSTR pszFaceName
  );

  petzold->hDlg=hDlg;
DVWRITELN("hDlg:%p hDc:%p",petzold->hDlg,petzold->hDc);
  petzold->hDc=GetDC(hDlg);
DVWRITELN("hDlg:%p hDc:%p",petzold->hDlg,petzold->hDc);

  if (!petzold->hDc) {
    DMESSAGE("getting hDc");
    goto e_hdc;
  }

  SelectObject(
    petzold->hDc,                   // HDC     hdc,
    hFont             // HGDIOBJ h
  );
#else // ] [
  petzold->hDc=GetDC(NULL);

  if (!petzold->hDc) {
    DMESSAGE("getting hDc");
    goto e_hdc;
  }
#endif // ]

#if ! defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
  petzold->font.hCur=CreateFontW(
#if 0 // [
    point_size,                     // int    cHeight,
#else // ] [
    -GetDeviceCaps(petzold->hDc,LOGPIXELSY)/6,
                                    // int    cHeight,
#endif // ]
    0,                              // int    cWidth,
    0,                              // int    cEscapement,
    0,                              // int    cOrientation,
    FW_DONTCARE,                    // int    cWeight,
    FALSE,                          // DWORD  bItalic,
    FALSE,                          // DWORD  bUnderline,
    FALSE,                          // DWORD  bStrikeOut,
    ANSI_CHARSET,                   // DWORD  iCharSet,
    OUT_DEFAULT_PRECIS,             // DWORD  iOutPrecision,
    CLIP_DEFAULT_PRECIS,            // DWORD  iClipPrecision,
    DEFAULT_QUALITY,                // DWORD  iQuality,
    DEFAULT_PITCH|FF_SWISS,         // DWORD  iPitchAndFamily,
    typeface                        // LPCSTR pszFaceName
  );

  if (!petzold->font.hCur) {
    DMESSAGE("creating font");
    goto e_font_cur;
  }

  petzold->font.hPrev=SelectObject(
    petzold->hDc,                   // HDC     hdc,
    petzold->font.hCur              // HGDIOBJ h
  );

  if (!petzold->font.hPrev) {
    DMESSAGE("selecting font");
    goto e_font_prev;
  }
#endif // ]

  GetScaledWidths (petzold) ;

  return 0;
//cleanup:
#if ! defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
  SelectObject(
    petzold->hDc,                   // HDC     hdc,
    petzold->font.hPrev             // HGDIOBJ h
  );
e_font_prev:
  DeleteObject(
    petzold->font.hCur              // HGDIOBJ ho
  );
e_font_cur:
#endif // ]
#if defined (PIB_WIDGET_PETZOLD2_CREATE_COMPATIBLE) // [
  DeleteDC(petzold->hDc);
e_hdc:
#elif defined (PIB_WIDGET_INTITALIZER_ARRANGE) // ] [
  ReleaseDC(hDlg,petzold->hDc);
e_hdc:
#else // ] [
  ReleaseDC(NULL,petzold->hDc);
e_hdc:
#endif // ]
#if defined (PIB_WIDGET_PETZOLD2_CREATE_COMPATIBLE) // [
  DeleteObject(petzold->screen.hBitmap);
e_screen_hbitmap:
  DeleteDC(petzold->screen.hDc);
e_screen_hdc:
#endif // ]
  return -1;
}

void pib_widget_petzold_destroy(PibWidgetPetzold *petzold)
{
#if defined (PIB_WIDGET_PETZOLD2_CREATE_COMPATIBLE) // [
  SelectObject(
    petzold->hDc,                   // HDC     hdc,
    petzold->font.hPrev             // HGDIOBJ h
  );

  DeleteObject(
    petzold->font.hCur              // HGDIOBJ ho
  );

  DeleteDC(petzold->hDc);
  DeleteObject(petzold->screen.hBitmap);
  DeleteDC(petzold->screen.hDc);
#if defined (PIB_WIDGET_INTITALIZER_ARRANGE) // [
  ReleaseDC(petzold->hDlg,petzold->hDc);
#else // ] [
  ReleaseDC(NULL,petzold->hDc);
#endif // ]
#endif // ]
}
#endif // ]
