/*
 * pib_widget_realizer.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <pib_widget.h>

#if ! defined (PIB_WIDGET_REALIZER2) // [
typedef const struct _PibWidgetRealizerMap PibWidgetRealizerMap;

struct _PibWidgetRealizerMap {
  long *(*x)(PibWidgetAcceptor *acceptor);
  long *(*y)(PibWidgetAcceptor *acceptor);
  long *(*cx)(SIZE *size);
  long *(*cy)(SIZE *size);
  EPibWidgetMode *(*mx)(PibWidgetMode *mode);
  EPibWidgetMode *(*my)(PibWidgetMode *mode);
};

static PibWidgetRealizerMap pib_map_affine;
static PibWidgetRealizerMap pib_map_twisted;
static PibWidgetAcceptorVmt vmt;

//#define PIB_REALIZER_MM
//#define PIB_REALIZER_ISOTROPIC
void pib_widget_realizer_create(PibWidgetAcceptor *acceptor,
    const wchar_t *title, WORD point_size, const wchar_t *typeface,
    PibWidgetBuf *buf)
{
  memset(acceptor,0,sizeof *acceptor);
  acceptor->vmt=&vmt;
  acceptor->realizer.title=title;
  acceptor->realizer.point_size=point_size;
  acceptor->realizer.typeface=typeface;
  acceptor->realizer.buf=buf;
#if defined (PIB_WIDGET_PETZOLD2) // [
  if (pib_widget_petzold_create(&acceptor->realizer.petzold,point_size,
      typeface)<0) {
    DMESSAGE("creating petzold");
    return;
  }
#elif ! defined (PIB_WIDGET_EXTENT_POINT_SIZE) // ] [
  acceptor->realizer.hDc=GetDC(NULL);

  acceptor->realizer.font.hCur=CreateFontW(
    point_size,                     // int    cHeight,
    0,                              // int    cWidth,
    0,                              // int    cEscapement,
    0,                              // int    cOrientation,
    FW_DONTCARE,                    // int    cWeight,
    FALSE,                          // DWORD  bItalic,
    FALSE,                          // DWORD  bUnderline,
    FALSE,                          // DWORD  bStrikeOut,
    ANSI_CHARSET,                   // DWORD  iCharSet,
    OUT_DEFAULT_PRECIS,             // DWORD  iOutPrecision,
    CLIP_DEFAULT_PRECIS,            // DWORD  iClipPrecision,
    DEFAULT_QUALITY,                // DWORD  iQuality,
    DEFAULT_PITCH|FF_SWISS,         // DWORD  iPitchAndFamily,
    typeface                        // LPCSTR pszFaceName
  );

  acceptor->realizer.font.hPrev=SelectObject(
    acceptor->realizer.hDc,         // HDC     hdc,
    acceptor->realizer.font.hCur    // HGDIOBJ h
  );

#if defined (PIB_REALIZER_MM) // [
  /*
   * The SetMapMode function sets the mapping mode of the specified
   * device context. The mapping mode defines the unit of measure used
   * to transform page-space units into device-space units, and also
   * defines the orientation of the device's x and y axes.
   *
   * iMode:  The new mapping mode. This parameter can be one of the
   *   following values.
   *   MM_ANISOTROPIC:  Logical units are mapped to arbitrary units
   *     with arbitrarily scaled axes. Use the SetWindowExtEx and
   *     SetViewportExtEx functions to specify the units, orientation,
   *     and scaling.
   *   MM_HIENGLISH:  Each logical unit is mapped to 0.001 inch. Positive
   *     x is to the right; positive y is up.
   *   MM_HIMETRIC:  Each logical unit is mapped to 0.01 millimeter.
   *     Positive x is to the right; positive y is up.
   *   MM_ISOTROPIC:  Logical units are mapped to arbitrary units with
   *     equally scaled axes; that is, one unit along the x-axis is equal
   *     to one unit along the y-axis. Use the SetWindowExtEx and
   *     SetViewportExtEx functions to specify the units and the orientation
   *     of the axes. Graphics device interface (GDI) makes adjustments as
   *     necessary to ensure the x and y units remain the same size (When
   *     the window extent is set, the viewport will be adjusted to keep
   *     the units isotropic).
   *   MM_LOENGLISH:  Each logical unit is mapped to 0.01 inch. Positive x
   *     is to the right; positive y is up.
   *   MM_LOMETRIC:  Each logical unit is mapped to 0.1 millimeter. Positive
   *     x is to the right; positive y is up.
   *   MM_TEXT:  Each logical unit is mapped to one device pixel. Positive x
   *     is to the right; positive y is down.
   *   MM_TWIPS:  Each logical unit is mapped to one twentieth of a printer's
   *     point (1/1440 inch, also called a twip). Positive x is to the right;
   *     positive y is up.
   */
  SetMapMode(
    acceptor->realizer.hDc,         // HDC hdc,
#if defined (PIB_REALIZER_ISOTROPIC) // [
    MM_ISOTROPIC                    // int iMode
#else // ] [
    MM_ANISOTROPIC                  // int iMode
#endif // ]
  );

  /*
   * First of all I'm setting the logical units to be used to draw to the
   * window to 2000 by 2000, and then making these units relate to 100 by
   * 100 pixels. So my text will be outputted at 100 pixels across and 100
   * pixels down in my client area. If I changed the TextOut call to paint
   * at 1000,1000 then the text will be drawn at 50,50 and changing it to
   * 4000,4000 will draw at 200,200.
   *
   * https://cboard.cprogramming.com/windows-programming/7925-setwindoworgex-setviewportorgex-setwindowextex-setviewportextex.html
   *
   * anyway, we let the origins of both, the window and the viewport,
   * be at (0,0).
   */
  SetWindowOrgEx(acceptor->realizer.hDc,0,0,NULL);
  SetViewportOrgEx(acceptor->realizer.hDc,0,0,NULL);

  /*
   * we still need some heuristics in order to make this work ...
   */
  switch (point_size) {
  case 8:
  case 9:
    SetWindowExtEx(acceptor->realizer.hDc,100,100,NULL);
    SetViewportExtEx(acceptor->realizer.hDc,400,100,NULL);
    SetWindowExtEx(acceptor->realizer.hDc,1000,1000,NULL);
    SetViewportExtEx(acceptor->realizer.hDc,900,1000,NULL);
    break;
  case 10:
  case 11:
    SetWindowExtEx(acceptor->realizer.hDc,1000,1000,NULL);
    SetViewportExtEx(acceptor->realizer.hDc,880,1000,NULL);
    break;
  case 12:
    SetWindowExtEx(acceptor->realizer.hDc,1000,1000,NULL);
    SetViewportExtEx(acceptor->realizer.hDc,670,1000,NULL);
    break;
  default:
    SetWindowExtEx(acceptor->realizer.hDc,1000,1000,NULL);
    SetViewportExtEx(acceptor->realizer.hDc,1000,1000,NULL);
    break;
  }
#endif // ]
#endif // ]

  pib_widget_realizer_reset(acceptor);
}

#define PIB_WIDGET_REALIZER_OVERLAPPEDWINDOW
void pib_widget_realizer_reset(PibWidgetAcceptor *acceptor)
{
  enum {
    STYLE=DS_3DLOOK
#if 0 // [
				|DS_SETFONT
        |DS_SHELLFONT
#endif // ]
        |DS_3DLOOK
        |DS_CENTER
#if defined (PIB_WIDGET_REALIZER_OVERLAPPEDWINDOW) // [
        |WS_OVERLAPPEDWINDOW
#else // ] [
        |DS_MODALFRAME
#endif // ]
        |DS_SYSMODAL
        |WS_CAPTION
        |WS_VISIBLE
        |WS_POPUP
        |WS_SYSMENU
  };

  pib_widget_buf_head(
    acceptor->realizer.buf,         // WidgetBuf *buf,
    STYLE,                          // DWORD style,
    0,                              // DWORD dwExtendedStyle,
    0,                              // short x,
    0,                              // short y,
    acceptor->realizer.x,           // short cx,
    acceptor->realizer.y,           // short cy,
    NULL,                           // const wchar_t *menu,
    NULL,                           // const wchar_t *cls,
    acceptor->realizer.title,       // const wchar_t *title,
    acceptor->realizer.point_size,  // WORD point_size,
    acceptor->realizer.typeface     // const wchar_t *typeface
  );

  acceptor->realizer.bounds.cx=acceptor->realizer.x;
  acceptor->realizer.bounds.cy=acceptor->realizer.y;
  acceptor->realizer.x=0;
  acceptor->realizer.y=0;
}

static void pib_widget_realizer_destroy(PibWidgetAcceptor *acceptor)
{
#if defined (PIB_WIDGET_PETZOLD2) // [
  pib_widget_petzold_destroy(&acceptor->realizer.petzold);
#elif ! defined (PIB_WIDGET_EXTENT_POINT_SIZE) // [
  SelectObject(
    acceptor->realizer.hDc,         // HDC     hdc,
    acceptor->realizer.font.hPrev   // HGDIOBJ h
  );

  DeleteObject(
    acceptor->realizer.font.hCur    // HGDIOBJ ho
  );

  ReleaseDC(
    NULL,                           // HWND hWnd,
    acceptor->realizer.hDc          // HDC  hDC
  );
#endif // ]
}

#if ! defined (PIB_WIDGET_REALIZER_CORE_VMT) // [
static void pib_widget_realizer_dispatch_physical(PibWidgetAcceptor *acceptor,
    PibWidget *widget, DWORD style, wchar_t *cls)
{
  enum { STYLE=WS_VISIBLE|WS_TABSTOP|CBS_DROPDOWNLIST|CBS_HASSTRINGS };
  const wchar_t *title=PIB_WIDGET_PHYSICAL(widget)->title;
  SIZE *geometry=widget->vmt->geometry(widget);

  if (!acceptor->realizer.buf->str) {
#if defined (PIB_WIDGET_PETZOLD2) // [
    widget->vmt->extent(widget,&acceptor->realizer.petzold,geometry);
#elif defined (PIB_WIDGET_EXTENT_POINT_SIZE) // ] [
    widget->vmt->extent(widget,acceptor->realizer.point_size,geometry);
#else // ] [
    widget->vmt->extent(widget,acceptor->realizer.hDc,
        acceptor->realizer.point_size,geometry);
#endif // ]
    PIB_WIDGET_PHYSICAL(widget)->idc=acceptor->realizer.buf->idc;
  }

//DVWRITELN("!!! \"%s\" cy:%ld",widget->vmt->id,widget->vmt->geometry(widget)->cy);
  pib_widget_buf_item(
    acceptor->realizer.buf,         // WidgetBuf *buf,
    style,                          // DWORD style,
    WS_EX_LEFT,                     // DWORD dwExtendedStyle,
    acceptor->realizer.x,           // short x,
    acceptor->realizer.y,           // short y,
    geometry->cx,                   // short cx,
    geometry->cy,                   // short cy,
    cls,                            // wchar_t *cls,
    title                           // wchar_t *title
  );

  acceptor->realizer.x+=geometry->cx;
  acceptor->realizer.y+=geometry->cy;
}
#endif // ]

static void pib_widget_realizer_dispatch_combobox(PibWidgetAcceptor *acceptor,
    PibWidget *widget)
{
#if defined (PIB_WIDGET_REALIZER_CORE_VMT) // [
  acceptor->core.vmt->dispatch_combobox(acceptor,widget);
#else // ] [
  enum { STYLE=WS_VISIBLE|WS_TABSTOP|CBS_DROPDOWNLIST|CBS_HASSTRINGS };

//DVWRITELN("*** cy:%ld",widget->vmt->geometry(widget)->cy);
  pib_widget_realizer_dispatch_physical(acceptor,widget,STYLE,WC_COMBOBOXW);
#endif // ]
}

static void pib_widget_realizer_dispatch_treeview(PibWidgetAcceptor *acceptor,
    PibWidget *widget)
{
#if defined (PIB_WIDGET_REALIZER_CORE_VMT) // [
  acceptor->core.vmt->dispatch_treeview(acceptor,widget);
#else // ] [
  enum { STYLE=WS_VISIBLE|WS_TABSTOP|WS_BORDER|TVS_INFOTIP|TVS_NOSCROLL };

//DVWRITELN("--- cy:%ld",widget->vmt->geometry(widget)->cy);
  pib_widget_realizer_dispatch_physical(acceptor,widget,STYLE,WC_TREEVIEWW);
#endif // ]
}

static void pib_widget_realizer_dispatch_button(PibWidgetAcceptor *acceptor,
    PibWidget *widget)
{
#if defined (PIB_WIDGET_REALIZER_CORE_VMT) // [
  acceptor->core.vmt->dispatch_button(acceptor,widget);
#else // ] [
  enum { STYLE=WS_VISIBLE|WS_TABSTOP|BS_PUSHBUTTON };

//DVWRITELN("+++ cy:%ld",widget->vmt->geometry(widget)->cy);
  pib_widget_realizer_dispatch_physical(acceptor,widget,STYLE,WC_BUTTONW);
#endif // ]
}

////////
static int pib_widget_realizer_border(PibWidgetAcceptor *acceptor)
{
  int border=0;
  int tos=acceptor->realizer.stack.tos;
  PibWidgetNode *node;
  PibWidgetList *list;
  
  while (tos) {
    --tos;
    node=acceptor->realizer.stack.node[tos];
    list=node->vmt->list(PIB_WIDGET(node));

    if (!list)
      break;

    border+=PIB_WIDGET_BOX(list)->border;
  }

  return border;
}

static void pib_widget_realizer_push(PibWidgetAcceptor *acceptor,
    PibWidgetNode *node)
{
  enum {
    SIZE_TOTAL=sizeof acceptor->realizer.stack.node,
    SIZE_ELEM=sizeof acceptor->realizer.stack.node[0],
    NELEMS=SIZE_TOTAL/SIZE_ELEM,
  };

  if (NELEMS<=acceptor->realizer.stack.tos) {
    DMESSAGE("overflow");
    return;
  }

  acceptor->realizer.stack.node[acceptor->realizer.stack.tos]=node;
  ++acceptor->realizer.stack.tos;
}

static PibWidgetNode *pib_widget_realizer_pop(PibWidgetAcceptor *acceptor)
{
  if (!acceptor->realizer.stack.tos) {
    DMESSAGE("underflow");
    return NULL;
  }

  return acceptor->realizer.stack.node[--acceptor->realizer.stack.tos];
}

/*
 * we have two phases:
 * phase 1  is identified by "realizer.buf->str" still be NULL. it calculates
 *   the box's outer bounds.
 * phase 2  calculates box's childrens position within the box and actually
 *   realizes them.
 */
static void pib_widget_realizer_dispatch_box(PibWidgetAcceptor *acceptor,
    PibWidget *widget, PibWidgetRealizerMap *map)
{
  PibWidgetList *list=widget->vmt->list(widget);
  SIZE *bounds UNUSED=list?list->vmt->geometry(PIB_WIDGET(list))
      :&acceptor->realizer.bounds;
  int realizing=acceptor->realizer.buf->str?1:0;
  SIZE *geometry=widget->vmt->geometry(widget);
  long border=PIB_WIDGET_BOX(widget)->border;
  long delta=PIB_WIDGET_BOX(widget)->delta;
  PibWidgetNode *cur;
  long x,y,cx,dy;

  if (!realizing) {
    // initialize the box's size, it's without border.
    *map->cx(geometry)=0;
    *map->cy(geometry)=0;
    dy=0;
  }
  else if (PIB_WIDGET_BOX(widget)->nvar) {
    dy=*map->cy(bounds)-*map->cy(geometry);
DVWRITELN("+++ \"%s\" DY:%ld",widget->vmt->id,dy);
    dy-=pib_widget_realizer_border(acceptor)<<1;
    *map->y(acceptor)-=dy>>1;
    dy/=PIB_WIDGET_BOX(widget)->nvar;
  }
  else
    dy=0;

  *map->x(acceptor)+=border;
  *map->y(acceptor)+=border;
  // we need to remember "y".
  y=*map->y(acceptor);

  for (cur=PIB_WIDGET_LIST(widget)->head;cur;cur=cur->next) {
    // remember "x".
    x=*map->x(acceptor); // {

    if (realizing) {
      // when realizing the box's bounds (cx,cy) are already known. anyway,
      // still we need to position the child within the box in front of
      // calling accept().
      cx=*map->cx(widget->vmt->geometry(widget));
      cx-=*map->cx(cur->vmt->geometry(PIB_WIDGET(cur)));

      switch (*map->mx(&cur->mode)) {
      case PIB_WIDGET_NONE:
        break;
      case PIB_WIDGET_BEGIN:
        break;
      case PIB_WIDGET_CENTER:
        *map->x(acceptor)+=cx>>1;
        break;
      case PIB_WIDGET_END:
        *map->x(acceptor)+=cx;
        break;
      default:
        break;
      }

      switch (*map->my(&cur->mode)) {
      case PIB_WIDGET_NONE:
        break;
      case PIB_WIDGET_BEGIN:
        break;
      case PIB_WIDGET_CENTER:
#if 0 // [
        *map->y(acceptor)
            +=(dy-*map->cy(cur->vmt->geometry(PIB_WIDGET(cur))))>>1;
#else // ] [
        *map->y(acceptor)+=(dy>>1);
#endif // ]
        break;
      case PIB_WIDGET_END:
        *map->y(acceptor)+=dy;
        break;
      default:
        DMESSAGE("something went wrong");
        break;
      }
    }

    ////////
    pib_widget_realizer_push(acceptor,cur);
    cur->vmt->accept(PIB_WIDGET(cur),acceptor);
    pib_widget_realizer_pop(acceptor);

    ////////
    if (realizing) {
      switch (*map->my(&cur->mode)) {
      case PIB_WIDGET_NONE:
        break;
      case PIB_WIDGET_BEGIN:
        *map->y(acceptor)+=dy;
        break;
      case PIB_WIDGET_CENTER:
#if 0 // [
        *map->y(acceptor)
            +=(dy-*map->cy(cur->vmt->geometry(PIB_WIDGET(cur))))>>1;
#else // ] [
        *map->y(acceptor)+=dy>>1;
#endif // ]
        break;
      case PIB_WIDGET_END:
        break;
      default:
        DMESSAGE("something went wrong");
        break;
      }
    }
    else {
      // we just have to consider (cx,cy). (x,y) is already advanced by
      // the call to "accept()" - indeed we make use of it.
      //
      // in x-direction, we take into account just the maximum.
      cx=*map->x(acceptor)-x;

      if (*map->cx(geometry)<cx)
        *map->cx(geometry)=cx;

      // in y-direction, the box is growing by each step.
      *map->cy(geometry)=*map->y(acceptor)-y;

      switch (*map->my(&cur->mode)) {
      case PIB_WIDGET_NONE:
        break;
      case PIB_WIDGET_BEGIN:
      case PIB_WIDGET_CENTER:
      case PIB_WIDGET_END:
        ++PIB_WIDGET_BOX(widget)->nvar;
        break;
      default:
        DMESSAGE("something went wrong");
        break;
      }
    }

    if (cur->next)
      *map->y(acceptor)+=delta;

    // restore "x".
    *map->x(acceptor)=x; // }
  }

  // "x" is still restored to it's initial value, hence in x-direction we
  // need to add "cx".
  *map->x(acceptor)+=*map->cx(geometry);
  *map->x(acceptor)+=border;
  *map->y(acceptor)+=border;
}

static void pib_widget_realizer_dispatch_hbox(PibWidgetAcceptor *acceptor,
    PibWidget *widget)
{
  pib_widget_realizer_dispatch_box(acceptor,widget,&pib_map_twisted);
}

static void pib_widget_realizer_dispatch_vbox(PibWidgetAcceptor *acceptor,
    PibWidget *widget)
{
  pib_widget_realizer_dispatch_box(acceptor,widget,&pib_map_affine);
}

////////
static PibWidgetAcceptorVmt vmt={
  .id="PIB_REALIZER",
  .destroy=pib_widget_realizer_destroy,
  .dispatch_combobox=pib_widget_realizer_dispatch_combobox,
  .dispatch_treeview=pib_widget_realizer_dispatch_treeview,
  .dispatch_button=pib_widget_realizer_dispatch_button,
  .dispatch_hbox=pib_widget_realizer_dispatch_hbox,
  .dispatch_vbox=pib_widget_realizer_dispatch_vbox,
};

////////
static long *_pib_map_affine_x(PibWidgetAcceptor *acceptor)
{
  return &acceptor->realizer.x;
}

static long *_pib_map_affine_y(PibWidgetAcceptor *acceptor)
{
  return &acceptor->realizer.y;
}

static long *_pib_map_affine_cx(SIZE *size)
{
  return &size->cx;
}

static long *_pib_map_affine_cy(SIZE *size)
{
  return &size->cy;
}

static EPibWidgetMode *_pib_map_affine_mx(PibWidgetMode *mode)
{
  return &mode->x;
}

static EPibWidgetMode *_pib_map_affine_my(PibWidgetMode *mode)
{
  return &mode->y;
}

static PibWidgetRealizerMap pib_map_affine={
  .x=_pib_map_affine_x,
  .y=_pib_map_affine_y,
  .cx=_pib_map_affine_cx,
  .cy=_pib_map_affine_cy,
  .mx=_pib_map_affine_mx,
  .my=_pib_map_affine_my,
};

////////
static long *_pib_map_twisted_x(PibWidgetAcceptor *acceptor)
{
  return &acceptor->realizer.y;
}

static long *_pib_map_twisted_y(PibWidgetAcceptor *acceptor)
{
  return &acceptor->realizer.x;
}

static long *_pib_map_twisted_cx(SIZE *size)
{
  return &size->cy;
}

static long *_pib_map_twisted_cy(SIZE *size)
{
  return &size->cx;
}

static EPibWidgetMode *_pib_map_twisted_mx(PibWidgetMode *mode)
{
  return &mode->y;
}

static EPibWidgetMode *_pib_map_twisted_my(PibWidgetMode *mode)
{
  return &mode->x;
}

static PibWidgetRealizerMap pib_map_twisted={
  .x=_pib_map_twisted_x,
  .y=_pib_map_twisted_y,
  .cx=_pib_map_twisted_cx,
  .cy=_pib_map_twisted_cy,
  .mx=_pib_map_twisted_mx,
  .my=_pib_map_twisted_my,
};
#endif // ]
