/*
 * pib_widget_realizer_core.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <pib_widget.h>

#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
#if defined (PIB_WIDGET_REALIZER_CORE_VMT) // [
static PibWidgetRealizerCoreVmt vmt;
#endif // ]

int pib_widget_realizer_core_create(PibWidgetRealizerCore *core, long border,
  WORD point_size, const wchar_t *type_face)
{
  static const wchar_t TEXT[]
      //=L"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
      //=L"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
      //=L"ABCDEFGHIJKLMNOPQRSTUVWXYZ";
      =L"0123456789";

  enum {
    TEXT_LEN=(sizeof TEXT)/(sizeof TEXT[0])-1
  };

  SIZE size;
  LOGFONTW lf;

#if defined (PIB_WIDGET_REALIZER_CORE_VMT) // [
  core->vmt=&vmt;
#endif // ]
#if defined (PIB_WIDGET_REALIZER2_GEOMETRY) // [
  core->geometry.cx=-1;
  core->geometry.cy=-1;
#endif // ]
  core->border=border;
  core->hDlg=NULL;

  /*
   * The CreateCompatibleDC function creates a memory device context (DC)
   * compatible with the specified device.
   *
   * When you no longer need the memory DC, call the DeleteDC function.
   */
  core->hDc=CreateCompatibleDC(NULL);

  if (!core->hDc) {
    DMESSAGE("getting device context");
    goto e_hdc;
  }

  /*
   * The CreateFontIndirect function creates a logical font that has the
   * specified characteristics. The font can subsequently be selected as
   * the current font for any device context.
   *
   * When you no longer need the font, call the DeleteObject function
   * to delete it.
   *
   * lfHeight > 0:  The font mapper transforms this value into device units
   *   and matches it against the cell height of the available fonts.
   * lfHeight < 0:  The font mapper transforms this value into device units
   *   and matches its absolute value against the character height of the
   *   available fonts.
   */
  memset(&lf,0,sizeof lf);
  lf.lfHeight=-point_size,
  wcscpy(lf.lfFaceName,type_face);
  core->font.hCur=CreateFontIndirectW(&lf);

  if (!core->font.hCur) {
    DMESSAGE("creating font");
    goto e_font_cur;
  }

  core->font.hPrev=(HFONT)SelectObject(core->hDc,core->font.hCur);

  if (!core->font.hPrev) {
    DMESSAGE("selecting font");
    goto e_font_prev;
  }

  /*
   * Q145994: HOWTO: Calculate Dialog Units When Not Using the System Font
   *   https://jeffpar.github.io/kbarchive/kb/145/Q145994/
   */
  GetTextExtentPoint32W(
    core->hDc,          // HDC     hdc,
    TEXT,               // LPCWSTR lpString,
    TEXT_LEN,           // int     c,
    &size               // LPSIZE  psizl
  );

#if 0 // [
  core->unit.cx=((double)size.cx/(TEXT_LEN>>1)+1)/2;
#endif // ]
#if 0 // [
  core->unit.cx=((double)size.cx/52+1)/2;
#endif // ]
  core->unit.cx=(double)size.cx/TEXT_LEN;
  core->unit.cy=point_size;

  return 0;
//cleanup:
  SelectObject(core->hDc,core->font.hPrev);
e_font_prev:
  DeleteObject(core->font.hCur);
e_font_cur:
  DeleteDC(core->hDc);
e_hdc:
  return -1;
}

#if defined (PIB_WIDGET_REALIZER_CORE_VMT) // [
void pib_widget_realizer_core_dispatch_box(PibWidgetAcceptor *acceptor,
    PibWidget *widget, PibWidgetRealizerMap *map)
{
  PibWidgetRealizerCore *core=&acceptor->core;
  //PibWidgetRealizer *realizer=&acceptor->realizer;
  SIZE *geometry=widget->vmt->geometry(widget);
  //int realizing=realizer->buf.head?1:0;
  int realizing=core->vmt->realizing(acceptor);
  long delta=PIB_WIDGET_BOX(widget)->delta;
  PibWidgetNode *cur;
  double x,y,cx;

  if (!realizing) {
    // initialize the box's size.
    *map->cx(geometry)=0;
    *map->cy(geometry)=0;
  }

  // remember "y".
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
  y=*map->ty(&core->turtle);
#else // ] [
  y=*map->ty(&realizer->turtle);
#endif // ]

  for (cur=PIB_WIDGET_LIST(widget)->head;cur;cur=cur->next) {
    // remember "x".
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
    x=*map->tx(&core->turtle); // {
#else // ] [
    x=*map->tx(&realizer->turtle); // {
#endif // ]

    if (realizing) {
      cx=*map->cx(widget->vmt->geometry(widget));
      cx-=*map->cx(cur->vmt->geometry(PIB_WIDGET(cur)));

      switch (*map->mx(&cur->mode)) {
      case PIB_WIDGET_NONE:
        break;
      case PIB_WIDGET_BEGIN:
        break;
      case PIB_WIDGET_CENTER:
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
        *map->tx(&core->turtle)+=cx/2.0;
#else // ] [
        *map->tx(&realizer->turtle)+=cx/2.0;
#endif // ]
        break;
      case PIB_WIDGET_END:
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
        *map->tx(&core->turtle)+=cx;
#else // ] [
        *map->tx(&realizer->turtle)+=cx;
#endif // ]
        break;
      default:
        break;
      }
    }

    cur->vmt->accept(PIB_WIDGET(cur),acceptor);

    if (realizing) {
      switch (*map->mx(&cur->mode)) {
      case PIB_WIDGET_NONE:
        break;
      case PIB_WIDGET_BEGIN:
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
        *map->tx(&core->turtle)+=cx;
#else // ] [
        *map->tx(&realizer->turtle)+=cx;
#endif // ]
        break;
      case PIB_WIDGET_CENTER:
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
        *map->tx(&core->turtle)+=cx/2.0;
#else // ] [
        *map->tx(&realizer->turtle)+=cx/2.0;
#endif // ]
        break;
      case PIB_WIDGET_END:
        break;
      default:
        DMESSAGE("something went wrong");
        break;
      }
    }

    // in x-direction, we take into account just the maximum.
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
    cx=*map->tx(&core->turtle)-x;
#else // ] [
    cx=*map->tx(&realizer->turtle)-x;
#endif // ]

    if (*map->cx(geometry)<cx)
      *map->cx(geometry)=cx+0.5;

    // in y-direction, the box is growing by each step.
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
    *map->cy(geometry)=(*map->ty(&core->turtle)-y)+0.5;
#else // ] [
    *map->cy(geometry)=(*map->ty(&realizer->turtle)-y)+0.5;
#endif // ]

#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
    if (cur->next)
      *map->ty(&core->turtle)+=delta;

    // restore "x".
    *map->tx(&core->turtle)=x; // }
#else // ] [
    if (cur->next)
      *map->ty(&realizer->turtle)+=delta;

    // restore "x".
    *map->tx(&realizer->turtle)=x; // }
#endif // ]
  }

  // "x" is still restored to it's initial value, hence in x-direction we
  // need to add "cx".
#if defined (PIB_WIDGET_REALIZER2_MORPH) // [
  *map->tx(&core->turtle)+=*map->cx(geometry);
#else // ] [
  *map->tx(&realizer->turtle)+=*map->cx(geometry);
#endif // ]
}

static void pib_widget_realizer_core_destroy(PibWidgetRealizerCore *core)
#else // ] [
void pib_widget_realizer_core_destroy(PibWidgetRealizerCore *core)
#endif // ]
{
  SelectObject(core->hDc,core->font.hPrev);
  DeleteObject(core->font.hCur);
  DeleteDC(core->hDc);
}

#if defined (PIB_WIDGET_REALIZER_CORE_VMT) // [
static int pib_widget_realizer_core_realizing(PibWidgetAcceptor *acceptor)
{
  return acceptor->realizer.head?1:0;
}

static void pib_widget_realizer_core_dispatch_physical(
    PibWidgetAcceptor *acceptor, PibWidget *widget, DWORD style, wchar_t *cls)
{
  PibWidgetRealizerCore *core=&acceptor->core;
  PibWidgetRealizer *realizer=&acceptor->realizer;
  const wchar_t *title=PIB_WIDGET_PHYSICAL(widget)->title;
  SIZE *geometry=widget->vmt->geometry(widget);

  if (!realizer->head) {
    widget->vmt->extent(widget,acceptor,geometry);
    PIB_WIDGET_PHYSICAL(widget)->idc=realizer->buf.idc;
  }

  pib_widget_buf_item(
    &realizer->buf,                 // WidgetBuf *buf,
    style,                          // DWORD style,
    WS_EX_LEFT,                     // DWORD dwExtendedStyle,
    core->turtle.x*4/core->unit.cx+0.5,
                                    // short x,
    core->turtle.y*8/core->unit.cy+0.5,
                                    // short y,
    (double)geometry->cx*4/core->unit.cx+0.5,
                                    // short cx,
    (double)geometry->cy*8/core->unit.cy+0.5,
                                    // short cy,
    cls,                            // wchar_t *cls,
    title                           // wchar_t *title
  );

  core->turtle.x+=geometry->cx;
  core->turtle.y+=geometry->cy;
}

static void pib_widget_realizer_core_dispatch_treeview(
    PibWidgetAcceptor *acceptor, PibWidget *widget)
{
  enum { STYLE=WS_VISIBLE|WS_TABSTOP|WS_BORDER|TVS_INFOTIP|TVS_NOSCROLL };

  pib_widget_realizer_core_dispatch_physical(acceptor,widget,STYLE,
      WC_TREEVIEWW);
}

static void pib_widget_realizer_core_dispatch_combobox(
    PibWidgetAcceptor *acceptor, PibWidget *widget)
{
  enum { STYLE=WS_VISIBLE|WS_TABSTOP|CBS_DROPDOWNLIST|CBS_HASSTRINGS };

  pib_widget_realizer_core_dispatch_physical(acceptor,widget,STYLE,
      WC_COMBOBOXW);
}

static void pib_widget_realizer_core_dispatch_button(
    PibWidgetAcceptor *acceptor, PibWidget *widget)
{
  enum { STYLE=WS_VISIBLE|WS_TABSTOP|BS_PUSHBUTTON };

  pib_widget_realizer_core_dispatch_physical(acceptor,widget,STYLE,
      WC_BUTTONW);
}

static PibWidgetRealizerCoreVmt vmt={
  .destroy=pib_widget_realizer_core_destroy,
  .realizing=pib_widget_realizer_core_realizing,
  .dispatch_treeview=pib_widget_realizer_core_dispatch_treeview,
  .dispatch_combobox=pib_widget_realizer_core_dispatch_combobox,
  .dispatch_button=pib_widget_realizer_core_dispatch_button,
};
#endif // ]
#endif // ]
