/*
 * yas.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <sys/stat.h>
#include <yas.h>

static void yas_property_mode_changed(const PibProperty *property,
    const PibProperties *properties, void *data);
static void yas_property_histogram_changed(const PibProperty *property,
    const PibProperties *properties, void *data);

//#define YAS_PROPERTY_DOUBLE_BROKEN
int yas_create(Yas *yas, WNDPROC wndproc)
{
  enum {
    NPROPERTIES=(sizeof yas->property)/(sizeof yas->property.min[0])
  }; 

  const wchar_t PLUGINSW[]=L"Plugins";

#if defined (PIB_GEN) // [
  winampGeneralPurposePlugin *gen=winampGetGeneralPurposePlugin();
  wchar_t *inidir=(wchar_t*)SendMessage(gen->hwndParent,WM_WA_IPC,0,
      IPC_GETINIDIRECTORYW);
#else // ] [
  wchar_t *inidir=(wchar_t*)SendMessage(plugin_hwnd_main(),WM_WA_IPC,0,
      IPC_GETINIDIRECTORYW);
#endif // ]
  size_t len =wcslen(inidir)
      +1u                 // '\\'
      +wcslen(GEN_YASW)
      +1u                 // '\\'
      +wcslen(PLUGINSW)
      +4u;                // .ini
  PibProperty *min=yas->property.min;
  PibProperty *max=min+NPROPERTIES;
  yas_rand_t size;

  ////////
  memset(yas,0,sizeof *yas);

  ////////
  yas->ini.path=malloc((len+1)*sizeof yas->ini.path[0]);

  if (!yas->ini.path) {
    DMESSAGE("allocating path");
    goto e_ini;
  }

  swprintf(yas->ini.path,len+1,L"%s\\%s\\%s.ini",
      inidir,PLUGINSW,GEN_YASW);

  // consolidated variables:
  yas->variables.cons.x=-1l;
  yas->variables.cons.y=-1l;
  yas->variables.cons.active=1;
  yas->variables.cons.wrap.cur=100.0;
  yas->variables.cons.wrap.max=100.0;
  yas->variables.cons.page=0;
  yas->variables.cons.play=1;
  yas->variables.cons.histogram=0;
  yas->variables.cons.mode=YAS_RAND_MODE_CRYPT,
  yas->variables.cons.progress.cur=0.0;
  yas->variables.cons.progress.max=100.0;

  // transient variables:
  yas->variables.trans=yas->variables.cons;

  // peristent properties:
  pib_property_long_create(&yas->property.x,L"geometry",L"x",
      offsetof(YasVariables,x));
  pib_property_long_create(&yas->property.y,L"geometry",L"y",
      offsetof(YasVariables,y));
  pib_property_int_create(&yas->property.page,L"common",L"page",
      offsetof(YasVariables,page));
  pib_property_int_create(&yas->property.active,L"common",L"active",
      offsetof(YasVariables,active));
#if ! defined (YAS_PROPERTY_DOUBLE_BROKEN) // [
  pib_property_double_create(&yas->property.wrap.cur,L"common",L"wrap",
      offsetof(YasVariables,wrap.cur));
#endif // ]
  pib_property_int_create(&yas->property.play,L"common",L"play",
      offsetof(YasVariables,play));
  pib_property_int_create_ex(&yas->property.mode,L"common",L"mode",
      offsetof(YasVariables,mode),yas_property_mode_changed,yas);
  pib_property_int_create_ex(&yas->property.histogram,L"common",L"histogram",
      offsetof(YasVariables,histogram),yas_property_histogram_changed,yas);

  // non-persistent properties:
#if defined (YAS_PROPERTY_DOUBLE_BROKEN) // [
  pib_property_double_create(&yas->property.wrap.cur,NULL,NULL,
      offsetof(YasVariables,wrap.cur));
#endif // ]
  pib_property_double_create(&yas->property.wrap.max,NULL,NULL,
      offsetof(YasVariables,wrap.max));
  pib_property_double_create(&yas->property.progress.cur,NULL,NULL,
      offsetof(YasVariables,progress.cur));
  pib_property_double_create(&yas->property.progress.max,NULL,NULL,
      offsetof(YasVariables,progress.max));

  ////////
DVWRITELNW("\"%s\"",yas->ini.path);
  if (pib_properties_create(&yas->properties,min,max,&yas->variables.cons,
      &yas->variables.trans,yas->ini.path)<0) {
    DMESSAGE("creating properties");
    goto e_properties;
  }

  ////////
  yas->yas_rand_create[YAS_RAND_MODE_CRYPT]=yas_rand_crypt_create;
  yas->yas_rand_create[YAS_RAND_MODE_WELL]=yas_rand_well_create;
  yas->yas_rand_create[YAS_RAND_MODE_SFMT]=yas_rand_sfmt_create;
  yas->yas_rand_create[YAS_RAND_MODE_LIBC]=yas_rand_libc_create;
  yas->yas_rand_create[YAS_RAND_MODE_MARSAGLIA_CMWC4096]
      =yas_rand_marsaglia_cmwc4096_create;

  ////////
  if (yas_histogram_create(&yas->hist)<0) {
    DMESSAGE("creating histogram");
    goto e_hist;
  }

  ////////
  if (pib_thread_common_create(&yas->sync,yas)<0) {
    DMESSAGE("creating sync thread");
    goto e_sync;
  }

  ////////
  yas->wndproc.wa=(WNDPROC)SetWindowLongW(
#if defined (PIB_GEN) // [
    gen->hwndParent,              // _In_ HWND hWnd,
#else // ] [
    plugin_hwnd_main(),           // _In_ HWND hWnd,
#endif // ]
    GWL_WNDPROC,                  // _In_ int  nIndex,
    (LONG)wndproc                 // _In_ LONG dwNewLong
  );

  size=(yas_rand_t)SendMessageW(
#if defined (PIB_GEN) // [
    gen->hwndParent,              // _In_ HWND    hWnd,
#else // ] [
    plugin_hwnd_main(),           // _In_ HWND    hWnd,
#endif // ]
    WM_WA_IPC,                    // _In_ UINT    Msg,
    0,                            // _In_ WPARAM  wParam,
    IPC_GETLISTLENGTH             // _In_ LPARAM  lParam
  );

  // we need to have the sync thread running in order to do the following:
  yas_reload(yas,size,yas_cons(yas)->mode);

  /*
   * we don't like to have the visualizer running when starting Winamp.
   *
   * TODO: should be moved to the input plugin.
   */
  PostMessageW(
#if defined (PIB_GEN) // [
    gen->hwndParent,            // _In_ HWND    hWnd,
#else // ] [
    plugin_hwnd_main(),         // _In_ HWND    hWnd,
#endif // ]
    WM_PIB_IPC,                 // _In_ UINT    Msg,
    YAS_INIT,                   // _In_ WPARAM  wParam,
    0                           // _In_ LPARAM  lParam
  );

  return 0;
//cleanup:
  pib_thread_common_destroy(&yas->sync);
e_sync:
  yas_histogram_destroy(&yas->hist);
e_hist:
  pib_properties_destroy(&yas->properties);
e_properties:
  free(yas->ini.path);
e_ini:
  return -1;
}

void yas_destroy(Yas *yas)
{
#if defined (PIB_GEN) // [
  winampGeneralPurposePlugin *gen=winampGetGeneralPurposePlugin();
#endif // ]

  pib_thread_common_destroy(&yas->sync);

  SetWindowLongW(
#if defined (PIB_GEN) // [
    gen->hwndParent,          // _In_ HWND hWnd,
#else // ] [
    plugin_hwnd_main(),       // _In_ HWND hWnd,
#endif // ]
    GWL_WNDPROC,              // _In_ int  nIndex,
    (LONG)yas->wndproc.wa  // _In_ LONG dwNewLong
  );

  yas_histogram_destroy(&yas->hist);
  pib_properties_destroy(&yas->properties);
  free(yas->ini.path);

  if (yas->state.path)
    free(yas->state.path);

  if (yas->shuffle)
    free(yas->shuffle);
}

static int yas_alloc(Yas *yas, yas_rand_t size)
{
  if (yas->size!=size) {
    ///////////////////////////////////////////////////////////////////////////
    if (yas->shuffle) {
      free(yas->shuffle);
      yas->shuffle=NULL;
    }

    yas->shuffle=calloc(size,sizeof yas->shuffle[0]);
      
    if (!yas->shuffle) {
      DMESSAGE("allocating shuffle");
      goto e_shuffle;
    }

    ///////////////////////////////////////////////////////////////////////////
    yas->size=size;
  }

  return 0;
e_shuffle:
  return -1;
}

static int yas_randomize(Yas *yas, yas_rand_t size, YasRandMode mode)
{
  int err=-1;
  YasRand rand;

  if (yas_alloc(yas,size)<0) {
    DMESSAGE("allocating shuffle");
    goto e_alloc;
  }

  if (yas->yas_rand_create[mode](&rand)<0) {
    DMESSAGE("creating randomizer");
    goto e_rand;
  }

#if defined (YAS_RAND_SHUFFLE_LOOP) // [
  if (yas_rand_shuffle_loop(&rand,yas->shuffle,size,4)<0) {
    DMESSAGE("shuffling");
    goto e_shuffle;
  }
#else // ] [
  if (yas_rand_shuffle(&rand,yas->shuffle,size)<0) {
    DMESSAGE("shuffling");
    goto e_shuffle;
  }
#endif // ]

  yas_cons(yas)->mode=mode;
  yas->cur=0;
#if defined (YAS_NEG) // [
  yas->neg=0;
#endif // ]
#if defined (YAS_PILOT) // [
  yas->pilot=0;
#endif // ]
  yas_sync_create(yas);
  err=0;
//cleanup:
e_shuffle:
  rand.vmt->destroy(&rand);
e_rand:
e_alloc:
  return err;
}

static int yas_load(Yas *yas, yas_rand_t size)
{
  int err=-1;
  FILE *f;
  yas_rand_t i;
  char patha[MAX_PATH];

  f=_wfopen(
    yas->state.path, // const wchar_t *filename,
    L"r"                // const wchar_t *mode
  );

  if (!f) {
    DVMESSAGE("opening file \"%s\"",PIB_WS2CS(yas->state.path,patha));
    goto e_file;
  }

  fscanf(f,"%u\n",&yas->cur);
#if defined (YAS_NEG) // [
  fscanf(f,"%u\n",&yas->neg);
#endif // ]
#if defined (YAS_PILOT) // [
  fscanf(f,"%u\n",&yas->pilot);
#endif // ]

  if (yas_alloc(yas,size)<0) {
    DMESSAGE("allocating shuffle");
    goto e_alloc;
  }

  for (i=0;i<yas->size;++i)
    fscanf(f,"%u\n",&yas->shuffle[i]);

  err=0;
//cleanup:
e_alloc:
  fclose(f);
e_file:
  return err;
}

#define IPC_GETPLUGINDIRECTORYW 1336
int yas_reload(Yas *yas, yas_rand_t size, YasRandMode mode)
{
  static const wchar_t PLUGINDIR[]=L"Plugins\\" GEN_YASW;

#if defined (PIB_GEN) // [
  winampGeneralPurposePlugin *gen=winampGetGeneralPurposePlugin();
  wchar_t *m3udir=(wchar_t*)SendMessage(gen->hwndParent,WM_WA_IPC,0,
      IPC_GETM3UDIRECTORYW);
#else // ] [
  wchar_t *m3udir=(wchar_t*)SendMessage(plugin_hwnd_main(),WM_WA_IPC,0,
      IPC_GETM3UDIRECTORYW);
#endif // ]
  yas_rand_t len=wcslen(m3udir)
      +1u                 // '\\'
      +wcslen(PLUGINDIR)
      +1u                 // '\\'
      +YAS_WIDTH          // %06u
      +1u                 // '-'
      +2u                 // %02d
      +4u;                // '.txt'
  int create=!yas->state.path
      ||yas->size!=size
      ||yas_cons(yas)->mode!=mode;
  struct _stat buf;
  int err;
  wchar_t *tok;
  yas_rand_t *cur;

  if (create) {
    ///////////////////////////////////////////////////////////////////////////
    if (yas->state.path) {
      free(yas->state.path);
      yas->state.path=NULL;
    }

    yas->state.path=malloc((len+1)*sizeof (yas->state.path[0]));

    if (!yas->state.path) {
      DMESSAGE("allocating path to state file");
      goto e_path;
    }

    swprintf(yas->state.path,len+1,
        L"%s\\%s\\%0" PIB_STRW(YAS_WIDTH) L"u-%02d.txt",
        m3udir,PLUGINDIR,size,mode);
  }

  /////////////////////////////////////////////////////////////////////////////
  yas_histogram_reset(&yas->hist,size);

  /////////////////////////////////////////////////////////////////////////////
  err=_wstat(
    yas->state.path,          // const wchar_t *path,
    &buf                      // struct _stat *buffer
  );

  if (err<0) {
    ///////////////////////////////////////////////////////////////////////////
    for (tok=wcstok(yas->state.path,L"\\");tok;tok=wcstok(NULL,L"\\")) {
      if (wcslen(yas->state.path)<len) {
        CreateDirectoryW(
          yas->state.path,    // LPCWSTR               lpPathName,
          NULL                // LPSECURITY_ATTRIBUTES lpSecurityAttributes
        );

        tok[wcslen(tok)]=L'\\';
      }
      else
        break;
    }

    if (yas_randomize(yas,size,mode)<0) {
      DMESSAGE("randomizing");
      goto e_rand;
    }
  }
  else if (yas_load(yas,size)<0) {
    DMESSAGE("loading");
    goto e_rand;
  }

#if defined (YAS_PILOT) // [
  for (cur=yas->shuffle;cur<=yas->shuffle+yas->pilot;++cur)
    yas_histogram_base_add(&yas->hist,*cur,size);
#else // ] [
  for (cur=yas->shuffle;cur<=yas->shuffle+yas->cur;++cur)
    yas_histogram_base_add(&yas->hist,*cur,size);
#endif // ]

  // IMPORTANT: make "mode" consolidated!
  yas_cons(yas)->mode=mode;
  yas_refresh(yas);

  return 0;
//cleanup:
e_rand:
  free(yas->state.path);
  yas->state.path=NULL;
e_path:
  return -1;
}

yas_rand_t yas_next(Yas *yas)
{
  ++yas->cur;

#if defined (YAS_PILOT) // [
  if (yas->cur==yas->size) {
    yas_randomize(yas,yas->size,yas_cons(yas)->mode);
    yas->pilot=0ul;
    yas_histogram_reset(&yas->hist,yas->hist.nbins);
    yas_histogram_incr_add(&yas->hist,yas->shuffle[yas->cur],yas->size);
  }
  else if (yas->pilot<yas->cur) {
    yas->pilot=yas->cur;
    yas_histogram_incr_add(&yas->hist,yas->shuffle[yas->cur],yas->size);
  }
#else // ] [
  yas_histogram_incr_add(&yas->hist,yas->shuffle[yas->cur],yas->size);
#endif // ]

  yas_sync_update(yas);
  yas_refresh(yas);

  return yas->shuffle[yas->cur];
}

yas_rand_t yas_prev(Yas *yas)
{
  --yas->cur;

  if (yas->cur==UINT32_MAX)
    yas->cur=yas->size-1;

  yas_sync_update(yas);
  yas_refresh(yas);

  return yas->shuffle[yas->cur];
}

static void yas_property_mode_changed(const PibProperty *property,
    const PibProperties *properties, void *data)
{
#if defined (PIB_GEN) // [
  winampGeneralPurposePlugin *gen=winampGetGeneralPurposePlugin();
#endif // ]
  Yas *yas=data;
  int mode=((YasVariables *)properties->trans)->mode;
  yas_rand_t size;

DWRITELN("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
  size=(yas_rand_t)SendMessageW(
#if defined (PIB_GEN) // [
    gen->hwndParent,                // _In_ HWND    hWnd,
#else // ] [
    plugin_hwnd_main(),             // _In_ HWND    hWnd,
#endif // ]
    WM_WA_IPC,                      // _In_ UINT    Msg,
    0,                              // _In_ WPARAM  wParam,
    IPC_GETLISTLENGTH               // _In_ LPARAM  lParam
  );

  if (!size) {
    DMESSAGE("something went wrong");
    return;
  }

  yas_reload(yas,size,mode);

  if (yas->shuffle) {
    CallWindowProcW(
      yas->wndproc.wa,              // _In_ WNDPROC lpPrevWndFunc,
#if defined (PIB_GEN) // [
      gen->hwndParent,              // _In_ HWND    hWnd,
#else // ] [
      plugin_hwnd_main(),           // _In_ HWND    hWnd,
#endif // ]
      WM_WA_IPC,                    // _In_ UINT    Msg,
      yas->shuffle[yas->cur],       // _In_ WPARAM  wParam,
      IPC_SETPLAYLISTPOS            // _In_ LPARAM  lParam
    );
  }
}

static void yas_property_histogram_changed(const PibProperty *property,
    const PibProperties *properties, void *data)
{
  Yas *yas=data;

  if (yas->form.main.vmt) {
DWRITELNW("+++++++++++++++++++++++++++++++++++++++++++++++++++");
    pib_form_refresh(&yas->form.main,yas->hDlg);
  }
}

void yas_sync_create(Yas *yas)
{
  YasThreadSyncRequestCreate *request=yas_thread_sync_request_create_new(yas);

  if (request)
    pib_thread_send(&yas->sync,&request->common);
}

void yas_sync_update(Yas *yas)
{
  YasThreadSyncRequestUpdate *request=yas_thread_sync_request_update_new(yas);

  if (request)
    pib_thread_send(&yas->sync,&request->common);
}
