/*
 * yas_form_configuration.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yas.h>
#include <yas/resource.h>

#if defined (DEBUG) // [
#define DID(ID) .id=PIB_STRA(ID) ":" PIB_STRA(__LINE__) ":mode_pairs",
#else // ] [
#define DID(ID)
#endif // ]

PibFormVmt yas_form_configuration_vmt;

static PibRadioGroupPair mode_pairs[]={
  { DID(IDC_RAND_CRYPT) .idc=IDC_RAND_CRYPT, .val=YAS_RAND_MODE_CRYPT },
  { DID(IDC_RAND_WELL)  .idc=IDC_RAND_WELL,  .val=YAS_RAND_MODE_WELL  },
  { DID(IDC_RAND_SFMT)  .idc=IDC_RAND_SFMT,  .val=YAS_RAND_MODE_SFMT  },
  { DID(IDC_RAND_LIBC)  .idc=IDC_RAND_LIBC,  .val=YAS_RAND_MODE_LIBC  },
  { DID(IDC_RAND_MARSAGLIA_CMWC4096)  .idc=IDC_RAND_MARSAGLIA_CMWC4096,
      .val=YAS_RAND_MODE_MARSAGLIA_CMWC4096  },
};

enum {
  MODE_NITEMS=(sizeof mode_pairs)/(sizeof mode_pairs[0]),
};

static void label_changed(PibControl *control, HWND hDlg)
{
  enum { SIZE=128 };
  HWND hLabel=GetDlgItem(hDlg,control->label.idc);
  PibForm *form=pib_control_form(control);
  double max=yas_form_trans(form)->wrap.max;
  double cur=yas_form_trans(form)->wrap.cur;
  wchar_t buf[SIZE];

  if (max)
    swprintf(buf,SIZE,L"%.0f%%",cur,max,100.0f*cur/max);
  else
    buf[0]=L'\0';

  SetWindowTextW(hLabel,buf);
}

enum {
  OFFS_SLIDER_WRAP_CUR,
  OFFS_STATIC_WRAP_CUR,
  OFFS_CHECK_BOX_PLAY,
  OFFS_RADIO_GROUP_MODE,
#if ! defined (YAS_SWAP_HISTOGRAM) // [
  OFFS_CHECK_BOX_HISTOGRAM,
#endif // ]
  OFFS_CHECK_BOX_ACTIVE,
};

static void slider_changed(PibControl *control, HWND hDlg)
{
  PibForm *form;

  form=pib_control_form(control);
  pib_control_refresh(form->min[OFFS_STATIC_WRAP_CUR],hDlg);
}

static PibControlConfig yas_form_configuration_config[]={
  { pib_control_slider_new,YAS_PROPERTY_WRAP_CUR,slider_changed,.slider={
    .idc=IDC_SLIDER_WRAP,
    .max_offs=YAS_PROPERTY_WRAP_MAX }},
  { pib_control_label_new,YAS_PROPERTY_WRAP_CUR,label_changed,.label={
    .idc=IDC_STATIC_WRAP }},
  { pib_control_check_box_new,YAS_PROPERTY_PLAY,.check_box={
    .idc=IDC_CHECK_BOX_PLAY }},
  { pib_control_radio_group_new,YAS_PROPERTY_MODE,.radio_group={
    .nitems=MODE_NITEMS, .min=mode_pairs }},
#if ! defined (YAS_SWAP_HISTOGRAM) // [
  { pib_control_check_box_new,YAS_PROPERTY_HISTOGRAM,.check_box={
    .idc=IDC_CHECK_BOX_HISTOGRAM }},
#endif // ]
  { pib_control_check_box_new,YAS_PROPERTY_ACTIVE,.check_box={
    .idc=IDC_CHECK_BOX_ACTIVE }},
};

enum {
  MASK_MODE_NITEMS=(sizeof yas_form_configuration_config)
      /(sizeof yas_form_configuration_config[0]),
};

static PibForm *yas_form_configuration_create(PibFormInit *init, HWND hDlg)
{
  PibForm *form=init->form;

  if (init->form_vmt!=&yas_form_configuration_vmt) {
    DMESSAGE("something went wrong");
    goto e_vmt;
  }

  if (pib_form_common_create(init,hDlg)<0) {
    DMESSAGE("creating form");
    goto e_form;
  }

  pib_form_refresh(init->form,hDlg);

  return form;
e_form:
e_vmt:
  return NULL;
}

BOOL CALLBACK yas_form_configuration_proc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  PibForm *form=(PibForm *)GetWindowLongPtrW(hDlg,GWLP_USERDATA);
  PibControl *control;

  switch(uMsg) {
  case WM_INITDIALOG:
    form=yas_form_configuration_create((PibFormInit *)lParam,hDlg);

    SetWindowLongPtrW(
      hDlg,               // HWND     hWnd,
      GWLP_USERDATA,      // int      nIndex,
      (LONG_PTR)form      // LONG_PTR dwNewLong
    );

    return TRUE;
  case WM_COMMAND:
    control=(PibControl *)GetWindowLongPtrW(GetDlgItem(hDlg,LOWORD(wParam)),
        GWLP_USERDATA);

    if (control)
      pib_control_notify(control,hDlg,uMsg,wParam,lParam);

    break;
  case WM_NOTIFY:
    control=(PibControl *)GetWindowLongPtrW(GetDlgItem(hDlg,LOWORD(wParam)),
        GWLP_USERDATA);

    if (control)
      pib_control_notify(control,hDlg,uMsg,wParam,lParam);

    break;
#if defined (SHOWWINDOW) // [
  case WM_SHOWWINDOW:
    if (!wParam)
      pib_form_refresh(form,hDlg);

    break;
#endif // ]
  case WM_PIB_IPC:
    switch (wParam) {
    case PIB_REFRESH:
      pib_form_refresh(form,hDlg);
      break;
    default:
      break;
    }

    break;
#if 0 // [
  case WM_LBUTTONDOWN:
  case WM_LBUTTONUP:
  case WM_MBUTTONDOWN:
  case WM_MBUTTONUP:
  case WM_RBUTTONDOWN:
  case WM_RBUTTONUP:
  case WM_XBUTTONDOWN:
  case WM_XBUTTONUP:
    return TRUE;
#endif // ]
  default:
    break;
  }

  return FALSE;
}

PibFormVmt yas_form_configuration_vmt={
#if defined (DEBUG) // [
  .id=__FILE__ ":" PIB_STRA(__LINE__) ":yas_form_configuration_vmt",
#endif // ]
  .label=L"Configuration",
  .idd=IDD_PAGE_CONFIGURATION,
  .ncontrols=(sizeof yas_form_configuration_config)
      /(sizeof yas_form_configuration_config[0]),
  .config=yas_form_configuration_config,
  .destroy=pib_form_common_destroy,
  .dlgproc=yas_form_configuration_proc,
};
