/*
 * yas_rand_marsaglia.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yas.h>
/*
 * Pseudo-random number generation algorithms
 *   https://mathoverflow.net/questions/29494/pseudo-random-number-generation-algorithms
 * Multiply-with-carry pseudorandom number generator
 *   https://en.wikipedia.org/wiki/Multiply-with-carry_pseudorandom_number_generator
 */
#if 0 // [
/*
 * https://groups.google.com/forum/#!msg/comp.lang.c/qZFQgKRCQGg/rmPkaRHqxOMJ
 *
 * Most RNGs work by keeping a certain number, say k,
 * of the most recently generated integers, then return
 * the next integer as a  function of those k.
 * The initial k integers, the seeds, are assumed to be
 * randomly chosen, usually 32-bits.
 * The period of the RNG is related to the number of
 * choices for seeds, usually 2^(32k), so to get longer
 * periods you need to increase k.
 * 
 * Probably the most common type has k=1, and needs a single seed,
 * with each new integer a function of the previous one.
 * An example is this congruential RNG, a form of which was the
 * system RNG in VAXs for many years:
 */
static unsigned long x=123456789; /* a random initial x to be */
                                  /* assigned by the calling program */
unsigned long cong(void)
{
  return (x=69069*x+362437);
}

/*
 * Simple, k=1, RNGs can perform fairly well in tests of randomness such as
 * those in the new version of Diehard,
 *            csis.hku.hk/~diehard
 * but experience has shown that better performances come
 * from RNGs with k's ranging from 4 or 5 to as much as 4097.
 * 
 * Here is an example with k=5, period about 2^160,
 * one of the fastest long period RNGs, returns more than
 * 120 million random 32-bit integers/second (1.8MHz CPU),
 * seems to pass all tests:
 */ 
static unsigned long x=123456789,y=362436069,z=521288629,w=88675123,v=886756453;
      /* replace defaults with five random seed values in calling program */
unsigned long xorshift(void)
{
  unsigned long t;

  t=(x^(x>>7));
  x=y;
  y=z;
  z=w;
  w=v;
  v=(v^(v<<6))^(t^(t<<13));
  
  return (y+y+1)*v;
}

/*
 * Another example has k=257,  period about 2^8222.
 * Uses a static array Q[256]  and an initial carry 'c',
 * the Q array filled with 256 random  32-bit integers
 * in the calling program and an initial carry c<809430660
 * for the multiply-with-carry operation.
 * It is very fast and seems to pass all tests.
 */
static unsigned long Q[256],c=362436;  /* choose random initial c<809430660 and
                                          256 random 32-bit integers for Q[] */

unsigned long MWC256(void)
{
  unsigned long long t,a=809430660LL;
  static unsigned char i=255;

  t=a*Q[++i]+c;
  c=(t>>32);

  return(Q[i]=t);
}


/*
 * The Mersenne Twister (check Google) is an excellent RNG,
 * with k=624. But it requires an elaborate C program and is
 * slower than many RNGs  that do as well in tests,
 * have comparable or longer periods and require
 * only a few lines of code.
 * 
 * Here is a complimentary-multiply-with-carry RNG
 * with k=4097 and a near-record period, more than
 * 10^33000 times as long as that of the Twister.
 * (2^131104 vs. 2^19937)
 */
static unsigned long Q[4096],c=362436; /* choose random initial
                                          c<809430660 and 4096 random
                                          32-bit integers for Q[] */
unsigned long CMWC4096(void)
{
  unsigned long long t, a=18782LL;
  static unsigned long i=4095;
  unsigned long x,r=0xfffffffe;

  i=(i+1)&4095;
  t=a*Q[i]+c;
  c=(t>>32); x=t+c; if(x<c){x++;c++;}

  return(Q[i]=r-x);
}
#endif // ]
static YasRandVmt yas_rand_marsaglia_cmwc4096_vmt;

int yas_rand_marsaglia_cmwc4096_create(YasRand *rand)
{
  int err=-1;
  HCRYPTPROV hProv;
  int bCode;

  rand->vmt=&yas_rand_marsaglia_cmwc4096_vmt;

  bCode=CryptAcquireContextW(
    &hProv,                     // HCRYPTPROV *phProv,
    NULL,                       // LPCSTR     szContainer,
    NULL,                       // LPCSTR     szProvider,
    PROV_RSA_FULL,              // DWORD      dwProvType,
    CRYPT_VERIFYCONTEXT         // DWORD      dwFlags
  );

  if (!bCode) {
    DMESSAGE("acquiring context");
    goto e_context;
  }

  bCode=CryptGenRandom(
    hProv,                      // HCRYPTPROV hProv,
    sizeof rand->marsaglia_cmwc4096.c,
                                // DWORD      dwLen,
    (BYTE *)&rand->marsaglia_cmwc4096.c
                                // BYTE       *pbBuffer
  );

  if (!bCode) {
    DMESSAGE("generating state");
    goto e_c;
  }

  bCode=CryptGenRandom(
    hProv,                      // HCRYPTPROV hProv,
    sizeof rand->marsaglia_cmwc4096.Q,
                                // DWORD      dwLen,
    (BYTE *)rand->marsaglia_cmwc4096.Q
                                // BYTE       *pbBuffer
  );

  if (!bCode) {
    DMESSAGE("generating state");
    goto e_q;
  }

  rand->marsaglia_cmwc4096.i=4095;
  err=0;
//cleanup:
e_q:
e_c:
  CryptReleaseContext(
    hProv,                      // HCRYPTPROV hProv,
    0ul                         // DWORD      dwFlags
  );
e_context:
  return err;
}

static void yas_rand_marsaglia_cmwc4096_destroy(YasRand *rand)
{
}

static int yas_rand_marsaglia_cmwc4096_next(YasRand *rand, yas_rand_t *result)
{
  uint64_t t,a=18782ull;
  yas_rand_t x,r=0xfffffffe;
  yas_rand_t *c=&rand->marsaglia_cmwc4096.c;
  yas_rand_t *Q=rand->marsaglia_cmwc4096.Q;
  uint16_t *i=&rand->marsaglia_cmwc4096.i;

  *i=(*i+1)&4095;
  t=a*Q[*i]+*c;
  *c=(t>>32);
  x=t+*c;

  if (x<*c) {
    ++x;
    ++*c;
  }

  *result=Q[*i]=r-x;

  return 0;
}

static YasRandVmt yas_rand_marsaglia_cmwc4096_vmt={
  .id=__FILE__,
  .max=UINT32_MAX,
  .destroy=yas_rand_marsaglia_cmwc4096_destroy,
  .next=yas_rand_marsaglia_cmwc4096_next,
};
/*
 * You will find several more CMWC RNGs and comments on
 * choice of seeds in the May 2003 Communications of the ACM.
 * 
 * George Marsaglia
 */
