/*
 * yas_rand_well.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <time.h>
#include <yas.h>

/*
 * Random Number Generation
 *   http://lomont.org/papers/2008/Lomont_PRNG_2008.pdf
 */
static YasRandVmt yas_rand_well_vmt;

#define YAS_RAND_WELL_SEED_CRYPT
int yas_rand_well_create(YasRand *rand)
{
  int err=-1;
#if defined (YAS_RAND_WELL_SEED_CRYPT) // [
  HCRYPTPROV hProv;
  int bCode;
#endif // ]

  rand->vmt=&yas_rand_well_vmt;

#if defined (YAS_RAND_WELL_SEED_CRYPT) // [
  bCode=CryptAcquireContextW(
    &hProv,                     // HCRYPTPROV *phProv,
    NULL,                       // LPCSTR     szContainer,
    NULL,                       // LPCSTR     szProvider,
    PROV_RSA_FULL,              // DWORD      dwProvType,
    CRYPT_VERIFYCONTEXT         // DWORD      dwFlags
  );

  if (!bCode) {
    DMESSAGE("acquiring context");
    goto e_context;
  }

DMARKLN();
  bCode=CryptGenRandom(
    hProv,                      // HCRYPTPROV hProv,
    sizeof rand->well.state,    // DWORD      dwLen,
    (BYTE *)rand->well.state    // BYTE       *pbBuffer
  );

  if (!bCode) {
    DMESSAGE("generating state");
    goto e_state;
  }

  rand->well.index=0;

  bCode=CryptGenRandom(
    hProv,                      // HCRYPTPROV hProv,
    1,                          // DWORD      dwLen,
    (BYTE *)&rand->well.index   // BYTE       *pbBuffer
  );

  if (!bCode) {
    DMESSAGE("generating index");
    goto e_index;
  }
#else // ] [
  memset(&rand->well,0,sizeof rand->well);
  rand->well.state[0]=time(NULL);
#endif // ]
  err=0;
//cleanup:
#if defined (YAS_RAND_WELL_SEED_CRYPT) // [
e_index:
e_state:
  CryptReleaseContext(
    hProv,                      // HCRYPTPROV hProv,
    0ul                         // DWORD      dwFlags
  );
e_context:
#endif // ]
  return err;
}

static void yas_rand_well_destroy(YasRand *rand)
{
}

static int yas_rand_well_next(YasRand *rand, yas_rand_t *result)
{
  yas_rand_t a, b, c, d;

  a=rand->well.state[rand->well.index];
  c=rand->well.state[(rand->well.index+13)&15];
  b=a^c^(a<<16)^(c<<15);
  c=rand->well.state[(rand->well.index+9)&15];
  c^=(c>>11);
  a=rand->well.state[rand->well.index] = b^c;
  d=a^((a<<5)&0xDA442D24UL);
  rand->well.index=(rand->well.index + 15)&15;
  a=rand->well.state[rand->well.index];
  *result=rand->well.state[rand->well.index]=a^b^d^(a<<2)^(b<<18)^(c<<28);

  return 0;
}

static YasRandVmt yas_rand_well_vmt={
  .id=__FILE__,
  .max=UINT32_MAX,
  .destroy=yas_rand_well_destroy,
  .next=yas_rand_well_next,
};
